# SPDX-License-Identifier: MIT
from ..common import *
from ...utils import align
from ...hw.dart import DART
from ...hw.isp import *

import atexit
from construct import *
import struct
import time

from .isp_chan import ISPChannelTable

class ISPSurface:
    def __init__(self, index, phys, iova, size, name):
        self.index = index
        self.phys = phys
        self.iova = iova
        self.size = size
        self.name = name

    def __str__(self):
        s = 'surface: [name: %s index: %d phys: 0x%x iova: 0x%x size: 0x%x]' % (self.name, self.index, self.phys, self.iova, self.size)
        return s

class ISPMemoryManager:
    def __init__(self, isp, start_index=0x8):
        self.isp = isp
        self.index = start_index
        self.map = {}
        self._stfu = False

    @property
    def stfu(self): self._stfu = True

    def log(self, *args):
        if (not self._stfu):
            if (args): print("ISP: MMGR:", *args)
            else: print()

    def alloc_size(self, size, name="NULL"):
        size = align(size, self.isp.page_size)
        phys = self.isp.u.memalign(self.isp.page_size, size)
        self.isp.p.memset32(phys, 0, size)
        iova = self.isp.dart.iomap(0, phys, size)

        surf = ISPSurface(self.index, phys, iova, size, name)
        self.map[self.index] = surf
        self.log(f'Mapped {surf}')
        self.index += 1
        return surf

    def free_surf(self, surf):
        if (surf.index not in self.map): return
        # this is python, we don't "free"
        self.log(f'Freed {surf}')

    def index_iova(self, iova):
        surf = [self.map[index] for index in self.map if self.map[index].iova == iova]
        if (len(surf) == 1): return surf[0]
        else: return None

    def dump(self):
        for index in self.map:
            surf = self.map[index]
            self.log(surf)


class ISPASC:
    def __init__(self, isp):
        self.isp = isp
        self.regs = self.isp.regs

    def boot(self):
        self.regs.ISP_ASC_CONTROL = 0x0
        self.regs.ISP_ASC_CONTROL = 0x10

    def shutdown(self):
        self.regs.ISP_ASC_CONTROL = 0x0

    def enable_interrupts(self):
        self.regs.ISP_IRQ_ENABLE = 0x0
        self.regs.ISP_IRQ_ENABLE = 0xf

    def disable_interrupts(self):
        self.regs.ISP_IRQ_ENABLE = 0x0

    def is_ready(self):
            status = self.regs.ISP_ASC_STATUS.val
            if (status & 0x3) == 0: # can't be 0x28, 0x2c
                    self.isp.log("ASC not in WFI; status: 0x%x" % status)
                    return False
            # normally 0x2a on first boot, 0x22 after
            self.isp.log("ASC in WFI; status: 0x%x" % status)
            return True

    def reset(self):
        self.regs.ISP_ASC_EDPRCR = 0x2  # I can't tell when this one's actually needed, so we'll do it every time

        self.regs.ISP_PMGR_0 = 0xff00ff
        self.regs.ISP_PMGR_1 = 0xff00ff
        self.regs.ISP_PMGR_2 = 0xff00ff
        self.regs.ISP_PMGR_3 = 0xff00ff

        self.regs.ISP_ASC_POWER_CYCLE_0 = 0xffffffff
        self.regs.ISP_ASC_POWER_CYCLE_1 = 0xffffffff
        self.regs.ISP_ASC_POWER_CYCLE_2 = 0xffffffff
        self.regs.ISP_ASC_POWER_CYCLE_3 = 0xffffffff
        self.regs.ISP_ASC_POWER_CYCLE_4 = 0xffffffff
        self.regs.ISP_ASC_POWER_CYCLE_5 = 0xffffffff

class ISP:
    def __init__(self, u):
        self.u = u
        self.p = u.proxy
        self.PAGE_SIZE = 0x4000
        self.page_size = self.PAGE_SIZE

        self.p.pmgr_adt_clocks_enable("/arm-io/isp")
        self.p.pmgr_adt_clocks_enable("/arm-io/dart-isp")
        self.dart = DART.from_adt(self.u, "/arm-io/dart-isp",
                iova_range=(0x3a28000, 0x20000000))  # +0x4000 extra heap
        self.dart.initialize()

        self.isp_base = u.adt["/arm-io/isp"].get_reg(0)[0]  # 0x22a000000
        self.pwr_base = u.adt["/arm-io/isp"].get_reg(1)[0]  # 0x23b700000
        self.isp_dart0_base = u.adt["/arm-io/dart-isp"].get_reg(0)[0]  # 0x22c0e8000
        self.isp_dart1_base = u.adt["/arm-io/dart-isp"].get_reg(1)[0]  # 0x22c0f4000
        self.isp_dart2_base = u.adt["/arm-io/dart-isp"].get_reg(2)[0]  # 0x22c0fc000

        self.regs = ISPRegs(backend=self.u, base=self.isp_base)
        self.ps = ISPPSRegs(backend=self.u, base=self.pwr_base)

        self.asc = ISPASC(self)
        self.mmger = ISPMemoryManager(self)
        self.table = None
        self.cmd_iova = 0x0
        self.frames = []
        self._stfu = False

    @property
    def stfu(self): self._stfu = True

    def log(self, *args):
        if (not self._stfu):
            if (args): print("ISP:", *args)
            else: print()

    def ioread(self, iova, size):
        data = self.dart.ioread(0, iova & 0xFFFFFFFFFF, size)
        return data

    def iowrite(self, iova, data):
        self.dart.iowrite(0, iova & 0xFFFFFFFFFF, data)

    def iomap(self, phys, size):
        iova = self.dart.iomap(phys, size)
        return iova

    def iomap_at(self, iova, phys, size):
        self.dart.iomap_at(0, iova & 0xFFFFFFFFFF, phys, size)

    def write_tunables(self):
            self.write_fabric_tunables()
            self.write_unk_power_tunables()
            self.write_dapf_tunables()
            self.write_dart_tunables()

    def write_fabric_tunables(self):
        p = self.p
        # Fabric tunables. Should go under tunables_static.c
        p.mask32(self.isp_base + 0x00, 0x10, 0x10)
        p.mask32(self.isp_base + 0x40, 0xffff, 0x50030)
        p.mask32(self.isp_base + 0x44, 0xffff, 0xa0040)
        p.mask32(self.isp_base + 0x400, 0x4, 0x40000001)
        p.mask32(self.isp_base + 0x600, 0x0, 0x1ffffff)
        p.mask32(self.isp_base + 0x738, 0x1ff01ff, 0x2)  # these 4 are used in power reset
        p.mask32(self.isp_base + 0x798, 0x1ff01ff, 0x300008)
        p.mask32(self.isp_base + 0x7f8, 0x1ff01ff, 0x880020)
        p.mask32(self.isp_base + 0x858, 0x1ff01ff, 0x200080)
        p.mask32(self.isp_base + 0x900, 0x1, 0x101)
        p.mask32(self.isp_base + 0x410, 0x100, 0x1100)
        p.mask32(self.isp_base + 0x420, 0x100, 0x1100)
        p.mask32(self.isp_base + 0x430, 0x100, 0x1100)
        p.mask32(self.isp_base + 0x8000, 0x0, 0x9)
        p.mask32(self.isp_base + 0x920, 0x0, 0x80)

    def write_unk_power_tunables(self):
        self.regs.ISP_POWER_UNK_0 = 0x1
        self.regs.ISP_POWER_UNK_1 = 0x80000000
        self.regs.ISP_POWER_UNK_2 = 0x80000000

    def write_dapf_tunables(self):
        p = self.p
        u = self.u
        dapf_cfg = getattr(u.adt['/arm-io/dart-isp'], 'filter-data-instance-0')
        dapf_base = u.adt['/arm-io/dart-isp'].reg[5].addr | 0x200000000
        offset = 0
        while offset < len(dapf_cfg):
                (start, end, _, r0h, r0l, _, r4) = struct.unpack_from('QQBBBBI', buffer=dapf_cfg, offset=offset)
                offset += 24
                p.write32(dapf_base + 0x4, r4)
                p.write32(dapf_base + 0x8, start & 0xFFFFFFFF)
                p.write32(dapf_base + 0xc, start >> 32) # macos does 32 bit writes, can we do 64?
                p.write32(dapf_base + 0x10, end & 0xFFFFFFFF)
                p.write32(dapf_base + 0x14, end >> 32)
                p.write32(dapf_base, (r0h << 4) | r0l)
                dapf_base += 0x40

    def write_dart_tunables(self):
        p = self.p
        isp_dart0_base = self.isp_dart0_base
        isp_dart1_base = self.isp_dart1_base
        isp_dart2_base = self.isp_dart2_base

        p.write32(isp_dart0_base + 0x100, 0x80)
        p.write32(isp_dart0_base + 0x13c, 0x100)

        p.write32(isp_dart1_base + 0xfc, 0x1)
        p.write32(isp_dart1_base + 0x200, self.dart.dart.regs.TTBR[0, 0].val)
        p.write32(isp_dart1_base + 0x2f0, 0x0)
        p.write32(isp_dart1_base + 0x34, 0xffffffff)
        p.write32(isp_dart1_base + 0x20, 0x100000)
        p.mask32(isp_dart1_base + 0x60, 0x10000, 0x80016100)
        p.mask32(isp_dart1_base + 0x68, 0x20202, 0xf0f0f)
        p.mask32(isp_dart1_base + 0x6c, 0x0, 0x80808)
        p.write32(isp_dart1_base + 0x100, 0x80)
        p.write32(isp_dart1_base + 0x13c, 0x20000)

        p.write32(isp_dart2_base + 0xfc, 0x1)
        p.write32(isp_dart2_base + 0x200, self.dart.dart.regs.TTBR[0, 0].val)
        p.write32(isp_dart2_base + 0x2f0, 0x0)
        p.write32(isp_dart2_base + 0x34, 0xffffffff)
        p.write32(isp_dart2_base + 0x20, 0x100000)
        p.mask32(isp_dart2_base + 0x60, 0x10000, 0x80016100)
        p.mask32(isp_dart2_base + 0x68, 0x20202, 0xf0f0f)
        p.mask32(isp_dart2_base + 0x6c, 0x0, 0x80808)
        p.write32(isp_dart2_base + 0x100, 0x80)
        p.write32(isp_dart2_base + 0x13c, 0x20000)

    def sync_ttbr(self):
        # Base ttbr is initialized after first iomap_at(), so copy it now
        self.p.write32(self.isp_dart1_base + 0x200, self.dart.dart.regs.TTBR[0, 0].val)
        self.p.write32(self.isp_dart2_base + 0x200, self.dart.dart.regs.TTBR[0, 0].val)

    def power_on(self):
        self.p.pmgr_adt_clocks_enable("/arm-io/isp")
        self.p.pmgr_adt_clocks_enable("/arm-io/dart-isp")

        # power domains, low -> high
        self.ps.ISP_PS_00 = 0xf
        self.ps.ISP_PS_08 = 0xf
        self.ps.ISP_PS_10 = 0xf
        self.ps.ISP_PS_18 = 0xf
        self.ps.ISP_PS_20 = 0xf
        self.ps.ISP_PS_28 = 0xf
        self.ps.ISP_PS_30 = 0xf
        self.ps.ISP_PS_38 = 0xf
        self.ps.ISP_PS_40 = 0xf
        self.ps.ISP_PS_48 = 0xf
        self.ps.ISP_PS_50 = 0xf
        self.ps.ISP_PS_58 = 0xf
        self.ps.ISP_PS_60 = 0xf

    def power_off(self):
        # power domains, high -> low
        self.ps.ISP_PS_60 = 0x0
        self.ps.ISP_PS_58 = 0x0
        self.ps.ISP_PS_50 = 0x0
        self.ps.ISP_PS_48 = 0x0
        self.ps.ISP_PS_40 = 0x0
        self.ps.ISP_PS_38 = 0x0
        self.ps.ISP_PS_30 = 0x0
        self.ps.ISP_PS_28 = 0x0
        self.ps.ISP_PS_20 = 0x0
        self.ps.ISP_PS_18 = 0x0
        self.ps.ISP_PS_10 = 0xf0017ff # intermediate state for the first three
        self.ps.ISP_PS_08 = 0xf0017ff
        self.ps.ISP_PS_00 = 0x7ff
        self.ps.ISP_PS_10 = 0x0 # now turn them off
        self.ps.ISP_PS_08 = 0x0
        self.ps.ISP_PS_00 = 0x0

        self.regs.ISP_DPE_UNK_1 = 0x103
        self.regs.ISP_DPE_UNK_0 = 0xc03  # self.p.mask32(0x22c504000, 0xc01, 0xc03)

    def initialize_firmware(self):

        # Stage0
        # =============================================================================
        """
        iova memory map

        0x0000000 - 0x09b4000; 0x09b4000: fw __TEXT
        0x09b4000 - 0x0dd0000; 0x041c000: fw __DATA
        0x0dd0000 - 0x1800000; 0x0a30000: internal heap fw uses for (fw) code execution
        0x1800000 - 0x1804000; 0x0004000: not mapped
        0x1804000 - 0x1820000; 0x001c000: ipc channels; see stage3()
        0x1820000 - 0x1824000; 0x0004000: not mapped
        0x1824000 - 0x3a24000; 0x2200000: extra heap requested by fw at boot
        """

        # text_phys = 0x8009e8000; text_iova = 0x000000; text_size = 0x9b4000;
        # data_phys = 0x8019b8000; data_iova = 0x9b4000; data_size = 0x41c000;
        (text_phys, text_iova, _, text_size, data_phys, data_iova, _, data_size) = struct.unpack('<QQQI4xQQQI4x', getattr(self.u.adt['/arm-io/isp'], 'segment-ranges'))
        self.iomap_at(text_iova, text_phys, text_size)
        self.iomap_at(data_iova, data_phys, data_size)

        heap_iova = data_iova + data_size # 0xdd0000
        heap_size = 0xa30000  # TODO I dont know if this is set
        heap_phys = self.u.heap.memalign(self.PAGE_SIZE, heap_size)
        self.p.memset32(heap_phys, 0, heap_size)
        self.iomap_at(heap_iova, heap_phys, heap_size)
        self.sync_ttbr()


        # Stage1
        # =============================================================================
        self.regs.ISP_SENSOR_CLOCK_0_EN = 0x1
        self.regs.ISP_GPIO_0 = 0x0
        self.regs.ISP_GPIO_1 = 0x0
        self.regs.ISP_GPIO_2 = 0x0
        self.regs.ISP_GPIO_3 = 0x0
        self.regs.ISP_GPIO_4 = 0x0
        self.regs.ISP_GPIO_5 = 0x0
        self.regs.ISP_GPIO_6 = 0x0
        self.regs.ISP_GPIO_7 = 0x0

        self.asc.boot()

        # Await ISP_GPIO_7 to 0x0 -> 0x8042006
        self.regs.ISP_GPIO_7 = 0x0  # Signal to fw
        for n in range(100):
                if (self.regs.ISP_GPIO_7.val == 0x8042006):
                        self.log('Got first magic number from firmware')
                        break
                time.sleep(0.01)
        assert((self.regs.ISP_GPIO_7.val == 0x8042006))

        ipc_chan_count = self.regs.ISP_GPIO_0.val  # 0x7
        ipc_args_offset = self.regs.ISP_GPIO_1.val  # 0xef40
        unk_2 = self.regs.ISP_GPIO_2.val  # 0x1
        extra_size = self.regs.ISP_GPIO_3.val  # 0x2200000; fw requested extra heap
        unk_4 = self.regs.ISP_GPIO_4.val  # 0x0
        assert((ipc_chan_count == 0x7))


        # Stage2
        # =============================================================================

        # Allocate IPC region
        ipc_iova = (heap_iova + heap_size) + 0x4000  # 0x1804000
        ipc_size = 0x1c000
        ipc_phys = self.u.heap.memalign(self.PAGE_SIZE, ipc_size)
        self.p.memset32(ipc_phys, 0, ipc_size)
        self.iomap_at(ipc_iova, ipc_phys, ipc_size)

        # Allocate extra heap requested by fw
        extra_iova = (ipc_iova + ipc_size) + 0x4000  # 0x1824000
        extra_phys = self.u.heap.memalign(self.PAGE_SIZE, extra_size)
        self.p.memset32(extra_phys, 0, extra_size)
        self.iomap_at(extra_iova, extra_phys, extra_size)

        bootargs_iova = (ipc_iova + ipc_args_offset) + 0x40  # (0x1824000 + 0xef40) + 0x40; 0x1812f80
        cmd_iova = (bootargs_iova + ISPIPCBootArgs.sizeof()) + 0x40  # (0x1812f80 + 0x180) + 0x40; 0x1813140
        self.cmd_iova = cmd_iova

        bootargs = ISPIPCBootArgs.build(dict(
                ipc_iova=ipc_iova,
                unk0=0x1800000,
                unk1=0xe800000,
                extra_iova=extra_iova,
                extra_size=extra_size,
                unk4=0x1,
                ipc_size=ipc_size,
                unk5=0x40,
                unk6=0x0,
                unk7=0x1,
                unk_iova=bootargs_iova+0x174,  # 0x18130f4; 0x180?
                unk9=0x3,
        ))
        self.iowrite(bootargs_iova, bootargs)

        self.regs.ISP_GPIO_0 = bootargs_iova
        self.regs.ISP_GPIO_1 = 0x0

        # Await ISP_GPIO_7 to 0xf7fbdff9 -> 0x8042006
        self.regs.ISP_GPIO_7 = 0xf7fbdff9  # Signal to fw
        for n in range(100):
                if (self.regs.ISP_GPIO_7.val == 0x8042006):
                        self.log('Got second magic number from firmware')
                        break
                time.sleep(0.01)
        assert((self.regs.ISP_GPIO_7.val == 0x8042006))


        # Stage3
        # =============================================================================

        table_iova = self.regs.ISP_GPIO_0.val  # 0x1804000
        unk_1 = self.regs.ISP_GPIO_1.val  # 0x0
        self.log("IPC channel description table at iova 0x%x" % table_iova)

        ipc_width = ISPIPCChanTableDescEntry.sizeof()  # 0x100
        table_data = self.ioread(table_iova, ipc_chan_count*ipc_width)
        description = []
        for n in range(ipc_chan_count):
                chan_data = table_data[n*ipc_width:(n+1)*ipc_width]
                desc = ISPIPCChanTableDescEntry.parse(chan_data)
                description.append(desc)
        self.table = ISPChannelTable(self, description)

        for chan in self.table.channels:
            if (chan.type == 0):  # IO, DEBUG, BUF_H2T
                for n in range(chan.num):
                    iova = chan.iova + (n * 0x40)
                    patch = struct.pack("<q", 0x1) + (struct.pack("<q", 0x0)*7)
                    self.iowrite(iova, patch)

        # Await ISP_GPIO_3 to 0x8042006 -> 0x0
        self.regs.ISP_GPIO_3 = 0x8042006  # Signal to fw
        for n in range(100):
                if (self.regs.ISP_GPIO_3.val == 0x0):
                        self.log('Got third magic number from firmware')
                        break
                time.sleep(0.01)
        assert((self.regs.ISP_GPIO_3.val == 0x0))

    def boot(self):
        self.power_on()
        self.write_tunables()

        self.asc.reset()
        assert(self.asc.is_ready())

        self.initialize_firmware()
        self.asc.enable_interrupts()
        atexit.register(self.shutdown)
        self.log("Firmware booted!")

        self.table.io.stfu
        self.table.bufh2t.stfu
        self.table.sharedmalloc.stfu
        self.table.sharedmalloc.handler()  # TODO don't hardcode this here
        self.log("Ready!")

    def shutdown(self):
        self.asc.disable_interrupts()
        self.asc.shutdown()
        self.power_off()



# SPDX-License-Identifier: MIT
import builtins
import sys

import __main__
from IPython import embed
from traitlets.config import get_config

c = get_config()
c.InteractiveShellEmbed.colors = "Linux"
from inspect import signature

from . import sysreg
from .proxy import *
from .proxyutils import *
from .utils import *

__all__ = ["run_ishell"]


cmd_list = {}
subcmd_list = {}
# Debug levels
DBL_NONE = 0
DBL_INFO = 1
DBL_TRACE = 2
DBL_DEBUG = 3
DBL_EDEBUG = 4

db_level = DBL_NONE


def debug_cmd(db=None):
    """Set debug level to integer %d(none)...%d(extreme debug)""" % (
        DBL_NONE,
        DBL_EDEBUG,
    )
    global db_level
    if db:
        db_level = db
    print("debug level=%d" % db_level)


def help_cmd(arg=None):
    if db_level >= DBL_DEBUG:
        print("arg=%s" % repr(arg))
    if arg:
        # cmd = arg.__qualname__
        if callable(arg):
            cmd = arg.__name__
        elif isinstance(arg, str):
            cmd = arg
        else:
            print("Unknown command: %s" % repr(arg))
            return
        if db_level >= DBL_DEBUG:
            print("cmd=%s" % repr(cmd))
        if cmd not in cmd_list:
            print("Undocumented command %s" % cmd)
            return
        hinfo = cmd_list[cmd]
        if isinstance(hinfo, str):
            print("%-10s : %s" % (cmd, hinfo))
            return
        if cmd in subcmd_list:
            clist = subcmd_list[cmd]
            aname = cmd
            if db_level >= DBL_DEBUG:
                print("subcmd_list[%s] = %s" % (repr(cmd), repr(clist)))
        else:
            print("command %s is not documented" % cmd)
            return
    else:
        clist = cmd_list
        aname = "top level"
        print("Note: To display a category's commands quote the name e.g. help('HV')")
    print("List of %s commands:" % aname)
    for cmd in clist.keys():
        hinfo = clist[cmd]
        if isinstance(hinfo, str):
            msg = hinfo.strip().split("\n", 1)[0]
        elif isinstance(hinfo, int):
            msg = "%s category - %d subcommands" % (cmd, hinfo)
        else:
            print("%s ?" % cmd)
            continue
        if len(cmd) <= 10:
            print("%-10s : %s" % (cmd, msg))
        else:
            print("%s:\n             %s" % (cmd, msg))


# commands is a dictionary for constructing the
# InteractiveConsole with. It adds in the callables
# in proxy utils iface and sysreg into commands
def run_ishell(commands, msg=""):
    saved_display = sys.displayhook
    try:

        def display(val):
            if isinstance(val, int) and not isinstance(val, bool):
                builtins._ = val
                print(hex(val))
            elif callable(val):
                val()
            else:
                saved_display(val)

        sys.displayhook = display

        # convenience
        commands["h"] = hex
        commands["sysreg"] = sysreg

        if "proxy" in commands and "p" not in commands:
            commands["p"] = commands["proxy"]
        if "utils" in commands and "u" not in commands:
            commands["u"] = commands["utils"]

        for obj_name in ("iface", "p", "u"):
            obj = commands.get(obj_name)
            obj_class = type(obj)
            if obj is None:
                continue

            for attr in dir(obj_class):
                if attr in commands or attr.startswith("_"):
                    continue

                member = getattr(obj_class, attr)
                if callable(member) and not isinstance(member, property):
                    cmd = getattr(obj, attr)
                    commands[attr] = cmd

        for attr in dir(sysreg):
            commands[attr] = getattr(sysreg, attr)

        commands["help"] = help_cmd
        commands["debug"] = debug_cmd
        for obj_name in commands.keys():
            obj = commands.get(obj_name)
            if obj is None or obj_name.startswith("_"):
                continue
            if callable(obj) and not isinstance(obj, property):
                try:
                    desc = obj_name + str(signature(obj))
                except:
                    continue
                qn = obj.__qualname__
                if qn.find(".") > 0:
                    a = qn.split(".")
                    if a[0] not in subcmd_list:
                        subcmd_list[a[0]] = {}
                    if a[0] not in cmd_list:
                        cmd_list[a[0]] = 1
                    else:
                        cmd_list[a[0]] += 1
                    clist = subcmd_list[a[0]]
                else:
                    clist = None
                if commands[obj_name].__doc__:
                    desc += " - " + commands[obj_name].__doc__
                    cmd_list[obj_name] = desc
                if isinstance(clist, dict):
                    clist[obj_name] = desc

        embed(config=c, extensions=["m1n1.ishell_ext"], header=msg, user_ns=commands)
    finally:
        sys.displayhook = saved_display


if __name__ == "__main__":
    from .setup import *

    commands = dict(__main__.__dict__)

    run_ishell(commands, msg="Have fun!")



# SPDX-License-Identifier: MIT

from enum import IntEnum
from m1n1.utils import *


class R_USB2PHY_USBCTL(Register32):
    MODE_HOST = 1
    MODE_ISOLATION = 2


class R_USB2PHY_CTL(Register32):
    RESET = 0
    PORT_RESET = 1
    APB_RESETN = 2
    SIDDQ = 3


class R_USB2PHY_SIG(Register32):
    VBUSDET_FORCE_VAL = 0
    VBUSDET_FORCE_EN = 1
    VBUSVLDEXT_FORCE_VAL = 2
    VBUSVLDEXT_FORCE_EN = 3
    MODE_HOST = 19, 12


class R_USB2PHY_MISCTUNE(Register32):
    APBCLK_GATE_OFF = 29
    REFCLK_GATE_OFF = 30


class Usb2PhyRegs(RegMap):
    USB2PHY_USBCTL = 0x00, R_USB2PHY_USBCTL
    USB2PHY_CTL = 0x04, R_USB2PHY_CTL
    USB2PHY_SIG = 0x08, R_USB2PHY_SIG
    USB2PHY_MISCTUNE = 0x1C, R_USB2PHY_MISCTUNE


class R_AUSPLL_DCO_EFUSE_SPARE(Register32):
    RODCO_ENCAP_EFUSE = 10, 9
    RODCO_BIAS_ADJUST_EFUSE = 14, 12


class R_AUSPLL_FRACN_CAN(Register32):
    DLL_START_CAPCODE = 18, 17


class R_AUSPLL_FSM_CTRL(Register32):
    APBREQ_OVSEL = 21, 13


class R_AUSPLL_CMD_OVERRIDE(Register32):
    APB_OVERRIDE = 28


class R_AUSPLL_CLKOUT_DTC_VREG(Register32):
    DTC_VREG_ADJUST = 16, 14


class R_AUS_COMMON_SHIM_BLK_VREG(Register32):
    VREG_TRIM = 6, 2


class R_CIO3PLL_CLK_CTRL(Register32):
    PCLK_EN = 1
    REFCLK_EN = 5


class R_CIO3PLL_DCO_NCTRL(Register32):
    DCO_COARSEBIN_EFUSE0 = 6, 0
    DCO_COARSEBIN_EFUSE1 = 23, 17


class R_CIO3PLL_FRACN_CAN(Register32):
    DLL_CAL_START_CAPCODE = 18, 17


class R_CIO3PLL_DTC_VREG(Register32):
    DTC_VREG_ADJUST = 16, 14


class E_ACIOPHY_CROSSBAR_PROTOCOL(IntEnum):
    USB4 = 0
    USB3 = 5
    USB3_DP = 8
    DP = 10


class R_ACIOPHY_CROSSBAR(Register32):
    PROTOCOL_SWAPPED = 0
    PROTOCOL = 4, 1, E_ACIOPHY_CROSSBAR_PROTOCOL
    DPMODE = 17, 5


class E_ACIOPHY_LANE_MODE(IntEnum):
    USB4 = 0
    USB3 = 1
    DP = 2
    OFF = 3


class R_ACIOPHY_LANE_MODE(Register32):
    RX0 = 2, 0, E_ACIOPHY_LANE_MODE
    TX0 = 5, 3, E_ACIOPHY_LANE_MODE
    RX1 = 8, 6, E_ACIOPHY_LANE_MODE
    TX1 = 11, 9, E_ACIOPHY_LANE_MODE


class R_ATCPHY_POWER(Register32):
    SLEEP_SMALL = 0
    SLEEP_BIG = 1
    CLAMP_EN = 2
    APB_RESET_N = 3
    PHY_RESET_N = 4


class R_ATCPHY_MISC(Register32):
    RESET_N = 0
    LANE_SWAP = 2


class R_AUSPMA_TX_SHM_TXA_CFG_MAIN_REG0(Register32):
    PMA_TXA_BYTECLK_RESET_SYNC_EN_OV = 2
    PMA_TXA_BYTECLK_RESET_SYNC_EN = 3
    PMA_TXA_BYTECLK_RESET_SYNC_CLR_OV = 4
    PMA_TXA_BYTECLK_RESET_SYNC_CLR = 5
    PMA_TXA_BYTECLK_RESET_SYNC_SEL_OV = 6


class R_AUSPMA_TX_SHM_TXA_CFG_MAIN_REG1(Register32):
    PMA_TXA_DIV2_EN_OV = 8
    PMA_TXA_DIV2_EN = 9
    PMA_TXA_DIV2_RESET_OV = 10
    PMA_TXA_DIV2_RESET = 11
    PMA_TXA_CLK_EN_OV = 22
    PMA_TXA_CLK_EN = 23


class R_AUSPMA_TX_SHM_TXA_IMP_REG0(Register32):
    PMA_TXA_CAL_CTRL_OV = 0
    PMA_TXA_CAL_CTRL = 18, 1
    PMA_TXA_CAL_CTRL_BASE_OV = 19
    PMA_TXA_CAL_CTRL_BASE = 23, 20
    PMA_TXA_HIZ_OV = 29
    PMA_TXA_HIZ = 30


class R_AUSPMA_TX_SHM_TXA_IMP_REG2(Register32):
    PMA_TXA_MARGIN_OV = 0
    PMA_TXA_MARGIN = 18, 1
    PMA_TXA_MARGIN_2R_OV = 19
    PMA_TXA_MARGIN_2R = 20


class R_AUSPMA_TX_SHM_TXA_IMP_REG3(Register32):
    PMA_TXA_MARGIN_POST_OV = 0
    PMA_TXA_MARGIN_POST = 10, 1
    PMA_TXA_MARGIN_POST_2R_OV = 11
    PMA_TXA_MARGIN_POST_2R = 12
    PMA_TXA_MARGIN_POST_4R_OV = 13
    PMA_TXA_MARGIN_POST_4R = 14
    PMA_TXA_MARGIN_PRE_OV = 15
    PMA_TXA_MARGIN_PRE = 21, 16
    PMA_TXA_MARGIN_PRE_2R_OV = 22
    PMA_TXA_MARGIN_PRE_2R = 23
    PMA_TXA_MARGIN_PRE_4R_OV = 24
    PMA_TXA_MARGIN_PRE_4R = 25


class R_AUSPMA_TX_SHM_TXA_LDOCLK(Register32):
    PMA_TXA_LDOCLK_BYPASS_SML_OV = 8
    PMA_TXA_LDOCLK_BYPASS_SML = 9
    PMA_TXA_LDOCLK_BYPASS_BIG_OV = 10
    PMA_TXA_LDOCLK_BYPASS_BIG = 11
    PMA_TXA_LDOCLK_EN_SML_OV = 12
    PMA_TXA_LDOCLK_EN_SML = 13
    PMA_TXA_LDOCLK_EN_BIG_OV = 14
    PMA_TXA_LDOCLK_EN_BIG = 15


class R_AUSPMA_RX_SHM_TJ_RXA_CTLE_CTRL0(Register32):
    PMA_RXA_TX_CLK_EN = 20
    PMA_RXA_TX_CLK_EN_OV = 21


class R_AUSPMA_RX_SHM_TJ_RXA_AFE_CTRL1(Register32):
    CLK_LANE_RX_DIV20_SYNC_RESET_N_OV = 29
    CLK_LANE_RX_DIV20_SYNC_RESET_N_VAL = 30


class R_AUSPMA_RX_SHM_TJ_RXA_DFE_CTRL10(Register32):
    PMA_RXA_DTVREG_ADJUST = 31, 27


class R_AUSPMA_RX_SHM_TJ_RXA_DFE_CTRL11(Register32):
    PMA_RXA_DTVREG_BIG_EN = 23
    PMA_RXA_DTVREG_BIG_EN_OV = 24
    PMA_RXA_DTVREG_SML_EN = 25
    PMA_RXA_DTVREG_SML_EN_OV = 26


class R_AUSPMA_RX_SHM_TJ_RXA_DFE_CTRL12(Register32):
    PMA_RXA_TX_BYTECLK_RESET_SYNC_CLR = 22
    PMA_RXA_TX_BYTECLK_RESET_SYNC_CLR_OV = 23
    PMA_RXA_TX_BYTECLK_RESET_SYNC_EN = 24
    PMA_RXA_TX_BYTECLK_RESET_SYNC_EN_OV = 25
    PMA_RXA_TX_HRCLK_SEL = 28
    PMA_RXA_TX_HRCLK_SEL_OV = 29
    PMA_RXA_TX_PBIAS_EN = 30
    PMA_RXA_TX_PBIAS_EN_OV = 31


class R_AUSPMA_RX_SHM_TJ_RXA_DFE_CTRL13(Register32):
    PMA_RXA_TX_PRE_EN = 0
    PMA_RXA_TX_PRE_EN_OV = 1
    PMA_RXA_TX_PST1_EN = 2
    PMA_RXA_TX_PST1_EN_OV = 3
    PMA_RXA_DTVREG_ADJUST_OV = 15


class R_AUSPMA_RX_SHM_TJ_RXA_SAVOS_CTRL16(Register32):
    PMA_RXA_RXTERM_EN = 21
    PMA_RXA_RXTERM_EN_OV = 22
    PMA_RXA_RXTERM_PULLUP_LEAK_EN = 23
    PMA_RXA_RXTERM_PULLUP_LEAK_EN_OV = 24
    PMA_RXA_TX_CAL_CODE = 29, 25
    PMA_RXA_TX_CAL_CODE_OV = 30


class R_AUSPMA_RX_SHM_TJ_RXA_TX_CTRL17(Register32):
    PMA_RXA_TX_MARGIN = 19, 15
    PMA_RXA_TX_MARGIN_OV = 20
    PMA_RXA_TX_MARGIN_LSB = 21
    PMA_RXA_TX_MARGIN_LSB_OV = 22
    PMA_RXA_TX_MARGIN_P1 = 26, 23
    PMA_RXA_TX_MARGIN_P1_OV = 27
    PMA_RXA_TX_MARGIN_P1_LSB = 29, 28
    PMA_RXA_TX_MARGIN_P1_LSB_OV = 30


class R_AUSPMA_RX_SHM_TJ_RXA_TX_CTRL18(Register32):
    PMA_RXA_TX_P1_CODE = 3, 0
    PMA_RXA_TX_P1_CODE_OV = 4
    PMA_RXA_TX_P1_LSB_CODE = 6, 5
    PMA_RXA_TX_P1_LSB_CODE_OV = 7
    PMA_RXA_TX_MARGIN_PRE = 10, 8
    PMA_RXA_TX_MARGIN_PRE_OV = 11
    PMA_RXA_TX_MARGIN_PRE_LSB = 13, 12
    PMA_RXA_TX_MARGIN_PRE_LSB_OV = 14
    PMA_RXA_TX_PRE_LSB_CODE = 16, 15
    PMA_RXA_TX_PRE_LSB_CODE_OV = 17
    PMA_RXA_TX_PRE_CODE = 21, 18
    PMA_RXA_TX_PRE_CODE_OV = 22


class R_AUSPMA_RX_SHM_TJ_RXA_TERM_CTRL19(Register32):
    PMA_RXA_TX_TEST_EN = 21
    PMA_RXA_TX_TEST_EN_OV = 22
    PMA_RXA_TX_EN = 23
    PMA_RXA_TX_EN_OV = 24
    PMA_RXA_TX_CLK_DLY_CTRL_TAPGEN = 27, 25
    PMA_RXA_TX_CLK_DIV2_EN = 28
    PMA_RXA_TX_CLK_DIV2_EN_OV = 29
    PMA_RXA_TX_CLK_DIV2_RST = 30
    PMA_RXA_TX_CLK_DIV2_RST_OV = 31


class R_AUSPMA_RX_SHM_TJ_RXA_VREF_CTRL22(Register32):
    PMA_RXA_VREF_ADJUST_GRAY = 11, 7
    PMA_RXA_VREF_ADJUST_GRAY_OV = 12
    PMA_RXA_VREF_BIAS_SEL = 14, 13
    PMA_RXA_VREF_BIAS_SEL_OV = 15
    PMA_RXA_VREF_BOOST_EN = 16
    PMA_RXA_VREF_BOOST_EN_OV = 17
    PMA_RXA_VREF_EN = 18
    PMA_RXA_VREF_EN_OV = 19
    LPBKIN_RECOVERED_DATA = 29, 28
    PMA_RXA_TEST_RXLPBKDT_EN = 30
    PMA_RXA_TEST_RXLPBKDT_EN_OV = 31


class R_AUSPMA_RX_TOP_TJ_CFG_RX_TXMODE(Register32):
    RX_TXMODE = 0

class R_ACIOPHY_LANE_DP_CFG_BLK_TX_DP_CTRL0(Register32):
    DP_PMA_BYTECLK_RESET = 0
    DP_MAC_DIV20_CLK_SEL = 1
    DPTXPHY_PMA_LANE_RESET_N = 2
    DPTXPHY_PMA_LANE_RESET_N_OV = 3
    DPTX_PCLK1_SELECT = 6, 4
    DPTX_PCLK2_SELECT = 9, 7
    DPRX_PCLK_SELECT = 12, 10
    DPTX_PCLK1_ENABLE = 13
    DPTX_PCLK2_ENABLE = 14
    DPRX_PCLK_ENABLE = 15


class AtcPhyRegs(RegMap):
    ACIOPHY_CFG0 = 0x08, Register32
    ACIOPHY_LANE_MODE = 0x48, R_ACIOPHY_LANE_MODE
    ACIOPHY_CROSSBAR = 0x4C, R_ACIOPHY_CROSSBAR
    ACIOPHY_BIST_EN = 0x84, Register32
    ACIOPHY_BIST_OV = 0x8C, Register32
    ACIOPHY_BIST_CFG0 = 0x90, Register32
    ACIOPHY_BIST_STAT = 0x9C, Register32
    ACIOPHY_BIST_RESET = 0xA8, Register32
    ACIOPHY_BIST_CFG1 = 0xAC, Register32
    ACIOPHY_SLEEP_CTRL = 0x1B0, Register32

    AUS_COMMON_SHIM_BLK_VREG = 0x0A04, R_AUS_COMMON_SHIM_BLK_VREG

    AUSPLL_FSM_CTRL = 0x1014, R_AUSPLL_FSM_CTRL
    AUSPLL_CMD_OVERRIDE = 0x2000, R_AUSPLL_CMD_OVERRIDE
    AUSPLL_CLKOUT_DTC_VREG = 0x2220, R_AUSPLL_CLKOUT_DTC_VREG
    AUSPLL_DCO_EFUSE_SPARE = 0x222C, R_AUSPLL_DCO_EFUSE_SPARE
    AUSPLL_FRACN_CAN = 0x22A4, R_AUSPLL_FRACN_CAN

    CIO3PLL_CLK_CTRL = 0x2A00, R_CIO3PLL_CLK_CTRL
    CIO3PLL_DTC_VREG = 0x2A20, R_CIO3PLL_DTC_VREG
    CIO3PLL_DCO_NCTRL = 0x2A38, R_CIO3PLL_DCO_NCTRL
    CIO3PLL_FRACN_CAN = 0x2AA4, R_CIO3PLL_FRACN_CAN

    ATCPHY_POWER_CTRL = 0x20000, R_ATCPHY_POWER
    ATCPHY_POWER_STAT = 0x20004, R_ATCPHY_POWER
    ATCPHY_MISC = 0x20008, R_ATCPHY_MISC

    ACIOPHY_LANE_DP_CFG_BLK_TX_DP_CTRL0 = 0x7000, R_ACIOPHY_LANE_DP_CFG_BLK_TX_DP_CTRL0
    DPPHY_UNK_1028 = 0x1028, Register32
    USB2PHY_AUX_CFG_BLK_AUX_POWER_DOWN_CONTROL_0 = 0x54000, Register32

    FABRIC_TUNABLES = irange(0x45000, 0x1000 // 4, 4), Register32

    LPDPTX_AUX_CFG_BLK = irange(0x50000, 0x1000 // 4, 4), Register32
    LPDPTX_AUX_CFG_BLK_AUX_CTRL = 0x50000, Register32
    LPDPTX_AUX_CFG_BLK_AUX_LDO_CTRL = 0x50008, Register32
    LPDPTX_AUX_CFG_BLK_AUX_MARGIN = 0x5000c, Register32
    LPDPTX_AUX_SHM_CFG_BLK_AUX_CTRL_REG0 = 0x50204, Register32
    LPDPTX_AUX_SHM_CFG_BLK_AUX_CTRL_REG1 = 0x50208, Register32

    LN0_AUSPMA_RX_TOP = irange(0x9000, 0x1000 // 4, 4), Register32
    LN0_AUSPMA_RX_TOP_TJ_CFG_RX_TXMODE = 0x90F0, R_AUSPMA_RX_TOP_TJ_CFG_RX_TXMODE

    LN0_AUSPMA_RX_EQ = irange(0xA000, 0x1000 // 4, 4), Register32

    LN0_AUSPMA_RX_SHM = irange(0xB000, 0x1000 // 4, 4), Register32
    LN0_AUSPMA_RX_SHM_TJ_RXA_CTLE_CTRL0 = 0xB000, R_AUSPMA_RX_SHM_TJ_RXA_CTLE_CTRL0
    LN0_AUSPMA_RX_SHM_TJ_RXA_AFE_CTRL1 = 0xB004, R_AUSPMA_RX_SHM_TJ_RXA_AFE_CTRL1
    LN0_AUSPMA_RX_SHM_TJ_RXA_UNK_CTRL2 = 0xB008, Register32
    LN0_AUSPMA_RX_SHM_TJ_RXA_UNK_CTRL3 = 0xB00C, Register32
    LN0_AUSPMA_RX_SHM_TJ_RXA_UNK_CTRL4 = 0xB010, Register32
    LN0_AUSPMA_RX_SHM_TJ_RXA_UNK_CTRL5 = 0xB014, Register32
    LN0_AUSPMA_RX_SHM_TJ_RXA_UNK_CTRL6 = 0xB018, Register32
    LN0_AUSPMA_RX_SHM_TJ_RXA_UNK_CTRL7 = 0xB01C, Register32
    LN0_AUSPMA_RX_SHM_TJ_RXA_UNK_CTRL8 = 0xB020, Register32
    LN0_AUSPMA_RX_SHM_TJ_RXA_UNK_CTRL9 = 0xB024, Register32
    LN0_AUSPMA_RX_SHM_TJ_RXA_DFE_CTRL10 = 0xB028, R_AUSPMA_RX_SHM_TJ_RXA_DFE_CTRL10
    LN0_AUSPMA_RX_SHM_TJ_RXA_DFE_CTRL11 = 0xB02C, R_AUSPMA_RX_SHM_TJ_RXA_DFE_CTRL11
    LN0_AUSPMA_RX_SHM_TJ_RXA_DFE_CTRL12 = 0xB030, R_AUSPMA_RX_SHM_TJ_RXA_DFE_CTRL12
    LN0_AUSPMA_RX_SHM_TJ_RXA_DFE_CTRL13 = 0xB034, R_AUSPMA_RX_SHM_TJ_RXA_DFE_CTRL13
    LN0_AUSPMA_RX_SHM_TJ_UNK_CTRL14A = 0xB038, Register32
    LN0_AUSPMA_RX_SHM_TJ_UNK_CTRL14B = 0xB03C, Register32
    LN0_AUSPMA_RX_SHM_TJ_UNK_CTRL15A = 0xB040, Register32
    LN0_AUSPMA_RX_SHM_TJ_UNK_CTRL15B = 0xB044, Register32
    LN0_AUSPMA_RX_SHM_TJ_RXA_SAVOS_CTRL16 = 0xB048, R_AUSPMA_RX_SHM_TJ_RXA_SAVOS_CTRL16
    LN0_AUSPMA_RX_SHM_TJ_RXA_TX_CTRL17 = 0xB04C, R_AUSPMA_RX_SHM_TJ_RXA_TX_CTRL17
    LN0_AUSPMA_RX_SHM_TJ_RXA_TX_CTRL18 = 0xB050, R_AUSPMA_RX_SHM_TJ_RXA_TX_CTRL18
    LN0_AUSPMA_RX_SHM_TJ_RXA_TERM_CTRL19 = 0xB054, R_AUSPMA_RX_SHM_TJ_RXA_TERM_CTRL19
    LN0_AUSPMA_RX_SHM_TJ_RXA_UNK_CTRL20 = 0xB058, Register32
    LN0_AUSPMA_RX_SHM_TJ_RXA_UNK_CTRL21 = 0xB05C, Register32
    LN0_AUSPMA_RX_SHM_TJ_RXA_VREF_CTRL22 = 0xB060, R_AUSPMA_RX_SHM_TJ_RXA_VREF_CTRL22

    LN0_AUSPMA_TX_TOP = irange(0xC000, 0x1000 // 4, 4), Register32

    LN0_AUSPMA_TX_SHM = irange(0xD000, 0x1000 // 4, 4), Register32
    LN0_AUSPMA_TX_SHM_TXA_CFG_MAIN_REG0 = 0xD000, R_AUSPMA_TX_SHM_TXA_CFG_MAIN_REG0
    LN0_AUSPMA_TX_SHM_TXA_CFG_MAIN_REG1 = 0xD004, R_AUSPMA_TX_SHM_TXA_CFG_MAIN_REG1
    LN0_AUSPMA_TX_SHM_TXA_IMP_REG0 = 0xD008, R_AUSPMA_TX_SHM_TXA_IMP_REG0
    LN0_AUSPMA_TX_SHM_TXA_IMP_REG1 = 0xD00C, Register32
    LN0_AUSPMA_TX_SHM_TXA_IMP_REG2 = 0xD010, R_AUSPMA_TX_SHM_TXA_IMP_REG2
    LN0_AUSPMA_TX_SHM_TXA_IMP_REG3 = 0xD014, R_AUSPMA_TX_SHM_TXA_IMP_REG3
    LN0_AUSPMA_TX_SHM_TXA_UNK_REG0 = 0xD018, Register32
    LN0_AUSPMA_TX_SHM_TXA_UNK_REG1 = 0xD01C, Register32
    LN0_AUSPMA_TX_SHM_TXA_UNK_REG2 = 0xD020, Register32
    LN0_AUSPMA_TX_SHM_TXA_LDOCLK = 0xD024, R_AUSPMA_TX_SHM_TXA_LDOCLK

    LN1_AUSPMA_RX_TOP = irange(0x10000, 0x1000 // 4, 4), Register32
    LN1_AUSPMA_RX_TOP_TJ_CFG_RX_TXMODE = 0x100F0, R_AUSPMA_RX_TOP_TJ_CFG_RX_TXMODE

    LN1_AUSPMA_RX_EQ = irange(0x11000, 0x1000 // 4, 4), Register32

    LN1_AUSPMA_RX_SHM = irange(0x12000, 0x1000 // 4, 4), Register32
    LN1_AUSPMA_RX_SHM_TJ_RXA_CTLE_CTRL0 = 0x12000, R_AUSPMA_RX_SHM_TJ_RXA_CTLE_CTRL0
    LN1_AUSPMA_RX_SHM_TJ_RXA_AFE_CTRL1 = 0x12004, R_AUSPMA_RX_SHM_TJ_RXA_AFE_CTRL1
    LN1_AUSPMA_RX_SHM_TJ_RXA_UNK_CTRL2 = 0x12008, Register32
    LN1_AUSPMA_RX_SHM_TJ_RXA_UNK_CTRL3 = 0x1200C, Register32
    LN1_AUSPMA_RX_SHM_TJ_RXA_UNK_CTRL4 = 0x12010, Register32
    LN1_AUSPMA_RX_SHM_TJ_RXA_UNK_CTRL5 = 0x12014, Register32
    LN1_AUSPMA_RX_SHM_TJ_RXA_UNK_CTRL6 = 0x12018, Register32
    LN1_AUSPMA_RX_SHM_TJ_RXA_UNK_CTRL7 = 0x1201C, Register32
    LN1_AUSPMA_RX_SHM_TJ_RXA_UNK_CTRL8 = 0x12020, Register32
    LN1_AUSPMA_RX_SHM_TJ_RXA_UNK_CTRL9 = 0x12024, Register32
    LN1_AUSPMA_RX_SHM_TJ_RXA_DFE_CTRL10 = 0x12028, R_AUSPMA_RX_SHM_TJ_RXA_DFE_CTRL10
    LN1_AUSPMA_RX_SHM_TJ_RXA_DFE_CTRL11 = 0x1202C, R_AUSPMA_RX_SHM_TJ_RXA_DFE_CTRL11
    LN1_AUSPMA_RX_SHM_TJ_RXA_DFE_CTRL12 = 0x12030, R_AUSPMA_RX_SHM_TJ_RXA_DFE_CTRL12
    LN1_AUSPMA_RX_SHM_TJ_RXA_DFE_CTRL13 = 0x12034, R_AUSPMA_RX_SHM_TJ_RXA_DFE_CTRL13
    LN1_AUSPMA_RX_SHM_TJ_UNK_CTRL14A = 0x12038, Register32
    LN1_AUSPMA_RX_SHM_TJ_UNK_CTRL14B = 0x1203C, Register32
    LN1_AUSPMA_RX_SHM_TJ_UNK_CTRL15A = 0x12040, Register32
    LN1_AUSPMA_RX_SHM_TJ_UNK_CTRL15B = 0x12044, Register32
    LN1_AUSPMA_RX_SHM_TJ_RXA_SAVOS_CTRL16 = 0x12048, R_AUSPMA_RX_SHM_TJ_RXA_SAVOS_CTRL16
    LN1_AUSPMA_RX_SHM_TJ_RXA_TX_CTRL17 = 0x1204C, R_AUSPMA_RX_SHM_TJ_RXA_TX_CTRL17
    LN1_AUSPMA_RX_SHM_TJ_RXA_TX_CTRL18 = 0x12050, R_AUSPMA_RX_SHM_TJ_RXA_TX_CTRL18
    LN1_AUSPMA_RX_SHM_TJ_RXA_TERM_CTRL19 = 0x12054, R_AUSPMA_RX_SHM_TJ_RXA_TERM_CTRL19
    LN1_AUSPMA_RX_SHM_TJ_RXA_UNK_CTRL20 = 0x12058, Register32
    LN1_AUSPMA_RX_SHM_TJ_RXA_UNK_CTRL21 = 0x1205C, Register32
    LN1_AUSPMA_RX_SHM_TJ_RXA_VREF_CTRL22 = 0x12060, R_AUSPMA_RX_SHM_TJ_RXA_VREF_CTRL22

    LN1_AUSPMA_TX_TOP = irange(0x13000, 0x1000 // 4, 4), Register32

    LN1_AUSPMA_TX_SHM = irange(0x14000, 0x1000 // 4, 4), Register32
    LN1_AUSPMA_TX_SHM_TXA_CFG_MAIN_REG0 = 0x14000, R_AUSPMA_TX_SHM_TXA_CFG_MAIN_REG0
    LN1_AUSPMA_TX_SHM_TXA_CFG_MAIN_REG1 = 0x14004, R_AUSPMA_TX_SHM_TXA_CFG_MAIN_REG1
    LN1_AUSPMA_TX_SHM_TXA_IMP_REG0 = 0x14008, R_AUSPMA_TX_SHM_TXA_IMP_REG0
    LN1_AUSPMA_TX_SHM_TXA_IMP_REG1 = 0x1400C, Register32
    LN1_AUSPMA_TX_SHM_TXA_IMP_REG2 = 0x14010, R_AUSPMA_TX_SHM_TXA_IMP_REG2
    LN1_AUSPMA_TX_SHM_TXA_IMP_REG3 = 0x14014, R_AUSPMA_TX_SHM_TXA_IMP_REG3
    LN1_AUSPMA_TX_SHM_TXA_UNK_REG0 = 0x14018, Register32
    LN1_AUSPMA_TX_SHM_TXA_UNK_REG1 = 0x1401C, Register32
    LN1_AUSPMA_TX_SHM_TXA_UNK_REG2 = 0x14020, Register32
    LN1_AUSPMA_TX_SHM_TXA_LDOCLK = 0x14024, R_AUSPMA_TX_SHM_TXA_LDOCLK

    ACIOPHY_TOP_TUNABLE_118 = 0x118, Register32
    ACIOPHY_TOP_TUNABLE_11c = 0x11C, Register32
    ACIOPHY_TOP_TUNABLE_124 = 0x124, Register32
    ACIOPHY_TOP_TUNABLE_a00 = 0xA00, Register32
    ACIOPHY_TOP_TUNABLE_808 = 0x808, Register32

    AUSPLL_TOP_FREQ_DESC_0A = 0x2080, Register32
    AUSPLL_TOP_FREQ_DESC_0B = 0x2084, Register32
    AUSPLL_TOP_FREQ_DESC_0C = 0x2088, Register32

    AUSPLL_TOP_TUNABLE_2094 = 0x2094, Register32
    AUSPLL_TOP_TUNABLE_20a0 = 0x20A0, Register32
    AUSPLL_TOP_TUNABLE_20ac = 0x20AC, Register32
    AUSPLL_TOP_TUNABLE_20b8 = 0x20B8, Register32

    CIO3PLL_TOP_TUNABLE_10 = 0x2810, Register32
    CIO3PLL_TOP_TUNABLE_88 = 0x2888, Register32
    CIO3PLL_TOP_TUNABLE_94 = 0x2894, Register32

    CIO3PLL_CORE_TUNABLE_1c = 0x2A1C, Register32
    CIO3PLL_CORE_TUNABLE_28 = 0x2A28, Register32
    CIO3PLL_CORE_TUNABLE_9c = 0x2A9C, Register32

    AUSPLL_CORE_TUNABLE_78 = 0x2278, Register32



# SPDX-License-Identifier: MIT

from m1n1.utils import RegMap
from m1n1.trace.i2c import I2CTracer, I2CRegMapTracer
from m1n1.hw.codecs import *

hv.p.hv_set_time_stealing(0, 1)

class SN012776Tracer(I2CRegMapTracer):
    REGMAP = SN012776Regs
    ADDRESSING = (1, 1)

class TAS5770Tracer(I2CRegMapTracer):
    REGMAP = TAS5770Regs
    ADDRESSING = (1, 1)

class CS42L84Tracer(I2CRegMapTracer):
    REGMAP = CS42L84Regs
    ADDRESSING = (0, 2)

class SSM3515Tracer(I2CRegMapTracer):
    REGMAP = SSM3515Regs
    ADDRESSING = (0, 1)

i2c_tracers = {}

for node in hv.adt["/arm-io"]:
    if not node.name.startswith("i2c"):
        continue

    n = int(node.name[3:])
    i2c_tracers[n] = bus = I2CTracer(hv, f"/arm-io/{node.name}")

    for devnode in node:
        if "compatible" not in devnode._properties: # thanks Apple
            continue

        dcls = {
            "audio-control,tas5770": TAS5770Tracer,
            "audio-control,sn012776": SN012776Tracer,
            "audio-control,cs42l84": CS42L84Tracer,
            "audio-control,ssm3515": SSM3515Tracer,
        }.get(devnode.compatible[0], None)
        if dcls:
            bus.add_device(
                devnode.reg[0] & 0xff,
                dcls(name=devnode.name)
            )

    bus.start()



# SPDX-License-Identifier: MIT
from construct import *

from ...utils import *
from ..asc import StandardASC
from ..afk.epic import *
from .dcpav import *

EOTF = "EOTF" / Enum(Int32ul,
    GAMMA_SDR = 1,
    GAMMA_HDR = 2,
)

Encoding = "Encoding" / Enum(Int32ul,
    RGB = 1,
    YCBCR_444 = 3,
    YCBCR_422 = 4,
    YCBCR_420 = 5,
)

Colorimetry = "Colorimetry" / Enum(Int32ul,
    BT601_709 = 1,
    BT2020 = 2,
    DCIP3 = 3,
)

Colorspace = "Colorspace" / Enum(Int32ul,
    SRGB = 1,
    Native = 2,
    BT2020 = 3,
)

SurfaceFormat = "SurfaceFormat" / Enum(Int32ul,
    BGRA = 1,
    BGRA2 = 2,
    RGBA = 3,
    w18p = 4,
    BGRA3 = 5,
    _444v = 6,
    _422v = 7,
    _420v = 8,
    w30r = 9,
    w40a = 10,
)

Transform = "Transform" / Enum(Int8ul,
    NONE = 0,
    XFLIP = 1,
    YFLIP = 2,
    ROT_90 = 3,
    ROT_180 = 4,
    ROT_270 = 5,
)

AddrFormat = "AddrFormat" / Enum(Int32ul,
    PLANAR = 1,
    TILED = 2,
    AGX = 3
)

TimingMode = Struct(
    "valid" / Bool(Int32ul),
    "width" / Int32ul,
    "height" / Int32ul,
    "fps_frac" / Int16ul,
    "fps_int" / Int16ul,
    Padding(8),
)

TimingModeList = Struct(
    "count" / Int32ul,
    "list" / GreedyRange(TimingMode),
)

ColorMode = Struct(
    "valid" / Bool(Int32ul),
    "colorimetry" / Colorimetry,
    "eotf" / EOTF,
    "encoding" / Encoding,
    "bpp" / Int32ul,
    "unk" / Int32ul,
)

ColorModeList = Struct(
    "count" / Int32ul,
    "list" / GreedyRange(ColorMode),
)

SwapInfo = Struct(
    "unk1" / Int32ul,
    "unk2" / Int32ul,
    "unk3" / Int32ul,
    "swap_id" / Int32ul,
    "unk5" / Int32ul,
)

IBootPlaneInfo = Struct(
    "unk1" / Default(Int32ul, 0),
    "addr" / Default(Int64ul, 0),
    "tile_size" / Default(Int32ul, 0),
    "stride" / Default(Int32ul, 0),
    "unk5" / Default(Int32ul, 0),
    "unk6" / Default(Int32ul, 0),
    "unk7" / Default(Int32ul, 0),
    "unk8" / Default(Int32ul, 0),
    "addr_format" / Default(AddrFormat, 0),
    "unk9" / Default(Int32ul, 0),
)

IBootLayerInfo = Struct(
    "planes" / Array(3, IBootPlaneInfo),
    "unk" / Default(Int32ul, 0),
    "plane_cnt" / Int32ul,
    "width" / Int32ul,
    "height" / Int32ul,
    "surface_fmt" / SurfaceFormat,
    "colorspace" / Colorspace,
    "eotf" / EOTF,
    "transform" / Transform,
    Padding(3)
)

SwapSetLayer = Struct(
    "unk" / Default(Int32ul, 0),
    "layer_id" / Int32ul,
    "layer_info" / IBootLayerInfo,
    "src_w" / Int32ul,
    "src_h" / Int32ul,
    "src_x" / Int32ul,
    "src_y" / Int32ul,
    "dst_w" / Int32ul,
    "dst_h" / Int32ul,
    "dst_x" / Int32ul,
    "dst_y" / Int32ul,
    "unk2" / Default(Int32ul, 0),
)

class DCPIBootService(EPICService):
    NAME = "disp0-service"
    SHORT = "disp0"

    def send_cmd(self, op, data=b'', replen=None):
        msg = struct.pack("<IIII", op, 16 + len(data), 0, 0) + data
        if replen is not None:
            replen += 8
        resp = super().send_cmd(0xc0, msg, replen)
        if not resp:
            return
        rcmd, rlen = struct.unpack("<II", resp[:8])
        return resp[8:rlen]

    def setPower(self, power):
        self.send_cmd(2, b"\x01" if power else b"\x00")

    def getModeCount(self):
        buf = self.send_cmd(3, b"", 12)
        hpd, timing_cnt, color_cnt = struct.unpack("<B3xII", buf)
        return bool(hpd), timing_cnt, color_cnt

    def getTimingModes(self):
        return TimingModeList.parse(self.send_cmd(4, replen=4096)).list

    def getColorModes(self):
        return ColorModeList.parse(self.send_cmd(5, replen=4096)).list

    def setMode(self, timing_mode, color_mode):
        data = TimingMode.build(timing_mode) + ColorMode.build(color_mode)
        self.send_cmd(6, data)

    def swapBegin(self):
        return SwapInfo.parse(self.send_cmd(15, replen=128))

    def swapSetLayer(self, layer_id, info, src_rect, dst_rect):
        data = Container()
        data.layer_id = layer_id
        data.layer_info = info
        data.src_w, data.src_h, data.src_x, data.src_y = src_rect
        data.dst_w, data.dst_h, data.dst_x, data.dst_y = dst_rect
        return self.send_cmd(16, SwapSetLayer.build(data), replen=128)

    def swapSetTimestamp(self):
        pass
        # 17

    def swapEnd(self):
        return self.send_cmd(18, b"\x00" * 12, replen=128)

    #def swapWait(self, swap_id):
        #buf = struct.pack("<IIII", 1, swap_id, 0, swap_id)
        #return self.send_cmd(19, buf, replen=128)

class DCPIBootEndpoint(EPICEndpoint):
    SHORT = "iboot"
    
    SERVICES = [
        DCPIBootService,
    ]


class DCPIBootClient(StandardASC):
    DVA_OFFSET = 0xf00000000

    ENDPOINTS = {
        0x20: AFKSystemEndpoint,
        0x23: DCPIBootEndpoint,
        0x24: DCPDPTXEndpoint,
        0x2a: DCPDPTXPortEndpoint,
        0x27: DCPAVDeviceEndpoint,
        0x28: DCPAVServiceEndpoint,
        0x29: DCPAVVideoEndpoint,
    }

    def __init__(self, u, asc_base, dart=None, disp_dart=None):
        super().__init__(u, asc_base, dart)
        self.disp_dart = disp_dart



# SPDX-License-Identifier: MIT

from m1n1 import asm
from m1n1.trace import Tracer
from m1n1.utils import *
from m1n1.proxy import *
from m1n1.sysreg import *
from m1n1.proxyutils import RegMonitor
from m1n1.trace.dart import DARTTracer
from m1n1.trace.asc import ASCTracer, EP, msg, msg_log, DIR
from m1n1.fw.pmp import *

#trace_device("/arm-io/pmgr", False)
#trace_device("/arm-io/jpeg0")
#trace_device("/arm-io/jpeg1")

#for reg in (0, 1, 2, 3, 4, 23):
    #addr, size = hv.adt["/arm-io/pmgr"].get_reg(reg)
    #hv.trace_range(irange(addr, 0x20000))

#hv.trace_range(irange(0x210e00000, 0x80000), read=False)
#hv.trace_range(irange(0x211e00000, 0x80000), read=False)

#hv.trace_range(irange(0x23b040000, 0x1000))
#hv.trace_range(irange(0x23b044000, 0x14000))

Tracer = Tracer._reloadcls()
ASCTracer = ASCTracer._reloadcls()

iomon = RegMonitor(hv.u, ascii=True)

def readmem_iova(addr, size):
    try:
        return dart_tracer.dart.ioread(0, addr, size)
    except Exception as e:
        print(e)
        return None

iomon.readmem = readmem_iova

class PMPEpTracer(EP):
    BASE_MESSAGE = PMPMessage

    def __init__(self, tracer, epid):
        super().__init__(tracer, epid)
        self.state.shmem_iova = None
        self.state.verbose = 1

    def start(self):
        self.add_mon()

    def add_mon(self):
        if self.state.shmem_iova:
            iomon.add(self.state.shmem_iova, 0x10000,
                      name=f"{self.name}.shmem@{self.state.shmem_iova:08x}", offset=0)

    @msg(1, DIR.TX, PMP_Configure)
    def Configure(self, msg):
        self.state.shmem_iova = msg.DVA
        self.add_mon()

class PMPTracer(ASCTracer):
    ENDPOINTS = {
        0x20: PMPEpTracer
    }

    def handle_msg(self, direction, r0, r1):
        super().handle_msg(direction, r0, r1)
        iomon.poll()

    def start(self, dart=None):
        super().start()
        # noisy doorbell
        self.trace(0x23bc34000, 4, TraceMode.OFF)

#dart_tracer = DARTTracer(hv, "/arm-io/dart-pmp", verbose=2)
#dart_tracer.start()

#pmp_tracer = PMPTracer(hv, "/arm-io/pmp", verbose=1)
#pmp_tracer.start(dart_tracer.dart)

class PMGRTracer(Tracer):
    IGNORED = set(["SPI1", "I2C2"])
    def __init__(self, hv):
        super().__init__(hv)
        self.dev = hv.adt["/arm-io/pmgr"]
        self.ignored_ranges = [
            (0x23b738004, 4), # ecpu state report
            (0x23b738008, 4), # pcpu state report
            (0x23d2b9000, 0x30),
            (0x23d2dc100, 4),
        ]
        self.build_table(hv)
        self.reg_cache = {}

    def hook_w(self, addr, val, width, **kwargs):
        self.hv.log(f"PMGR: W {addr:#x} <- {val:#x}")
        #print("-> ignored")
        super().hook_w(addr, val, width, **kwargs)

    def hook_r(self, addr, width, **kwargs):
        val = super().hook_r(addr, width, **kwargs)
        self.hv.log(f"PMGR: R {addr:#x} = {val:#x}")
        return val

    def evt_rw(self, evt):
        if not evt.flags.WRITE:
            self.reg_cache[evt.addr] = evt.data
        cb = self.ranges.lookup(evt.addr)
        cb[0](evt, *cb[1:])
        self.reg_cache[evt.addr] = evt.data

    def event_default(self, evt, start, name):
        t = "W" if evt.flags.WRITE else "R"
        m = "+" if evt.flags.MULTI else " "
        data = f"{evt.data:#x}"
        if evt.flags.WRITE:
            data = f"{evt.data:#x}"
            old = self.reg_cache.get(evt.addr, None)
            if old is not None:
                data = f"{old:#x} -> {evt.data:#x}"
        self.hv.log(f"[cpu{evt.flags.CPU}][0x{evt.pc:016x}] PMGR: {t}.{1<<evt.flags.WIDTH:<2}{m} " +
                    f"0x{evt.addr:x} ({name} + {evt.addr - start:#04x}) = {data}", show_cpu=False)

    def build_table(self, hv):
        self.ranges = ScalarRangeMap()
        self.state_regs = {}

        starts = {}
        for reg in (0, 1):
            addr, size = self.dev.get_reg(reg)
            self.ranges[addr:addr + size] = self.event_default, addr, f"reg[{reg}]"

        for i, ps in enumerate(self.dev.ps_regs):
            addr = self.dev.get_reg(ps.reg)[0] + ps.offset
            for idx in range(32):
                ps_addr = addr + idx * 8
                self.ranges[ps_addr:ps_addr + 8] = self.event_default, ps_addr, f"ps[{i}][{idx}]"

        for i, dev in enumerate(self.dev.devices):
            ps = self.dev.ps_regs[dev.psreg]
            if dev.psidx or dev.psreg:
                addr = self.dev.get_reg(ps.reg)[0] + ps.offset + dev.psidx * 8
                self.state_regs[addr] = dev.name
                if dev.name in self.IGNORED:
                    self.ignored_ranges.append((addr, 8))
                self.ranges[addr:addr + 8] = self.event_default, addr, f"{dev.name}.pstate"

    def start(self):
        self.hv.clear_tracers(self.ident)

        for reg in (0, 1):
            addr, size = self.dev.get_reg(reg)
            self.trace(addr, size, TraceMode.WSYNC, read=False)

        for ps in self.dev.ps_regs:
            addr = self.dev.get_reg(ps.reg)[0] + ps.offset
            self.trace(addr, 0x100, TraceMode.WSYNC)

        for lane in range(8):
            addr = 0x200200000 + 0x40000 * lane
            self.trace(addr, 0x40000, TraceMode.HOOK)
        #for reg in (23,):
            #addr, size = self.dev.get_reg(reg)
            #self.trace(addr, 0x20000, TraceMode.SYNC)
        for addr, size in self.ignored_ranges:
            self.trace(addr, size, TraceMode.OFF)

pmgr_tracer = PMGRTracer(hv)
pmgr_tracer.start()



# SPDX-License-Identifier: MIT
from ..utils import chexdump
from ..malloc import Heap
from construct.core import *
from ..fw.agx.channels import *
from ..fw.agx.cmdqueue import *
from ..fw.agx.microsequence import *
from ..hw.uat import MemoryAttr
from .object import *
import textwrap

class GPUContext:
    def __init__(self, agx):
        self.agx = agx
        self.uat = self.agx.uat
        self.u = self.agx.u
        self.p = self.agx.p
        self.verbose = False
        self.ctx = None

        #self.job_list = agx.kshared.new(JobList)
        #self.job_list.first_job = 0
        #self.job_list.last_head = self.job_list._addr # Empty list has self as last_head
        #self.job_list.unkptr_10 = 0
        #self.job_list.push()

        self.ttbr0_base = self.u.memalign(self.agx.PAGE_SIZE, self.agx.PAGE_SIZE)
        self.p.memset32(self.ttbr0_base, 0, self.agx.PAGE_SIZE)

        self.objects = {}

        # 32K VA pages since buffer manager needs that
        self.uobj = GPUAllocator(agx, "Userspace", 0x1600000000, 0x100000000, ctx=None,
                                 guard_pages=1,
                                 va_block=32768, nG=1, AP=0, PXN=1, UXN=1)

        self.gobj = GPUAllocator(agx, "GEM", 0x1500000000, 0x100000000, ctx=None,
                                 guard_pages=1, nG=1, AP=0, PXN=1, UXN=1)

        self.pipeline_base = 0x1100000000
        self.pipeline_size = 1 << 32
        self.pobj = GPUAllocator(agx, "Pipelines", self.pipeline_base + 0x10000, self.pipeline_size,
                                 ctx=None, guard_pages=1, nG=1, AP=0, PXN=1, UXN=1)

    def bind(self, ctx_id):
        self.ctx = ctx_id
        self.uobj.ctx = ctx_id
        self.gobj.ctx = ctx_id
        self.pobj.ctx = ctx_id
        self.uat.bind_context(ctx_id, self.ttbr0_base)
        self.thing = self.buf_at(0x6fffff8000, 0, 0x4000, "thing")

    def make_stream(self, base):
        return self.uat.iostream(self.ctx, base, recurse=False)

    def new_at(self, addr, objtype, name=None, track=True, **flags):
        obj = GPUObject(self, objtype)
        obj._stream = self.make_stream
        if name is not None:
            obj._name = name

        size_align = align_up(obj._size, self.agx.PAGE_SIZE)
        obj._addr = addr

        obj._paddr = self.agx.u.memalign(self.agx.PAGE_SIZE, size_align)
        #if isinstance(obj.val, ConstructClassBase):
            #obj.val._addr = obj._addr

        self.agx.log(f"[Context@{self.ctx}] Map {obj._name} size {obj._size:#x} @ {obj._addr:#x} ({obj._paddr:#x})")

        flags2 = {"AttrIndex": MemoryAttr.Shared}
        flags2.update(flags)
        obj._map_flags = flags2

        obj._size_align = size_align
        self.agx.uat.iomap_at(self.ctx, obj._addr, obj._paddr, size_align, **flags2)
        self.objects[obj._addr] = obj
        self.agx.reg_object(obj, track=track)

        return obj

    def buf_at(self, addr, is_pipeline, size, name=None, track=True):
        return self.new_at(addr, Bytes(size), name, track=track,
                           AttrIndex=MemoryAttr.Shared, PXN=1,
                           nG=1, AP=(1 if is_pipeline else 0))

    def load_blob(self, addr, is_pipeline, filename, track=True):
        data = open(filename, "rb").read()
        obj = self.new_at(addr, Bytes(len(data)), filename, track=track,
                          AttrIndex=MemoryAttr.Shared, PXN=1,
                          nG=1, AP=(1 if is_pipeline else 0))
        obj.val = data
        obj.push()

        return obj

    def free(self, obj):
        obj._dead = True
        self.agx.uat.iomap_at(self.ctx, obj._addr, 0, obj._size_align, VALID=0)
        del self.objects[obj._addr]
        self.agx.unreg_object(obj)

    def free_at(self, addr):
        self.free(self.objects[obj._addr])

class GPUWorkQueue:
    def __init__(self, agx, scheduler_context, job_list):
        self.agx = agx
        self.u = agx.u
        self.p = agx.p

        self.info = agx.kobj.new(CommandQueueInfo)

        self.pointers = agx.kshared.new(CommandQueuePointers).push()
        self.pmap = CommandQueuePointerMap(self.u, self.pointers._paddr)

        self.rb_size = self.pointers.rb_size
        self.ring = agx.kobj.new_buf(8 * self.rb_size, "GPUWorkQueue.RB")

        self.info.pointers = self.pointers
        self.info.rb_addr = self.ring._addr
        self.info.job_list = job_list
        self.info.gpu_buf_addr = agx.kobj.buf(0x2c18, "GPUWorkQueue.gpu_buf")
        self.info.gpu_context = scheduler_context
        self.info.push()

        self.wptr = 0
        self.first_time = True

        self.agx.uat.flush_dirty()

    def submit(self, work):
        work.push()

        self.p.write64(self.ring._paddr + 8 * self.wptr, work._addr)
        self.wptr = (self.wptr + 1) % self.rb_size
        self.agx.uat.flush_dirty()
        self.pmap.CPU_WPTR.val = self.wptr

    def wait_empty(self):
        while self.wptr != self.pmap.GPU_DONEPTR.val:
            self.agx.work()

class GPU3DWorkQueue(GPUWorkQueue):
    TYPE = 1

class GPUTAWorkQueue(GPUWorkQueue):
    TYPE = 0

class GPUMicroSequence:
    def __init__(self, agx):
        self.agx = agx
        self.off = 0
        self.ops = []
        self.obj = self.agx.kobj.new_buf(4096, "GPUMicroSequence", track=True)

    def cur_addr(self):
        return self.obj._addr + self.off

    def append(self, op):
        off = self.off
        self.ops.append(op)
        self.off += op.sizeof()
        return off

    def finalize(self):
        self.ops.append(EndCmd())
        self.size = sum(i.sizeof() for i in self.ops)
        self.obj.val = b"".join(i.build() for i in self.ops) + bytes(4096 - self.size)
        self.obj.push()
        return self.obj

    def dump(self):
        chexdump(self.agx.iface.readmem(self.obj._paddr, self.size))
        print(MicroSequence.parse_stream(self.agx.uat.iostream(0, self.obj._addr)))

    def __str__(self):
        s = f"GPUMicroSequence: {len(self.ops)} ops\n"
        for i, op in enumerate(self.ops):
            op_s = textwrap.indent(str(op), ' ' * 4)
            s += f"[{i:2}:{op.sizeof():#x}] = {op!s}\n"
        return s

class GPUBufferManager:
    def __init__(self, agx, context, blocks=8):
        self.agx = agx
        self.ctx = context

        self.block_ctl_obj = agx.kshared.new(BufferManagerBlockControl)
        self.block_ctl_obj.total = blocks
        self.block_ctl_obj.wptr = 0
        self.block_ctl_obj.unk = 0
        self.block_ctl = self.block_ctl_obj.push().regmap()

        self.counter_obj = agx.kshared.new(BufferManagerCounter)
        self.counter_obj.count = 0
        self.counter = self.counter_obj.push().regmap()

        self.misc_obj = agx.kshared.new(BufferManagerMisc)
        self.misc_obj.cpu_flag = 1
        self.misc = self.misc_obj.push().regmap()

        self.page_size = 0x8000
        self.pages_per_block = 4
        self.block_size = self.pages_per_block * self.page_size

        self.scene_max = 48

        self.page_list = context.uobj.new(Array(0x10000 // 4, Int32ul), "BM PageList", track=False)
        self.block_list = context.uobj.new(Array(0x8000 // 4, Int32ul), "BM BlockList", track=False)
        self.scene_list = agx.kgpurw.new(Array(self.scene_max, Int32ul), "BM SceneList", track=True)

        self.scene_list.val = [0] * 48
        self.scene_list.push()

        self.info = info = agx.kobj.new(BufferManagerInfo)
        info.page_list_addr = self.page_list._addr
        info.page_list_size = self.page_list._size
        info.page_count = self.block_ctl_obj.total * 4
        info.block_count = self.block_ctl_obj.total

        info.block_list_addr = self.block_list._addr
        info.block_ctl = self.block_ctl_obj
        info.last_page = info.page_count - 1
        info.block_size = self.block_size

        info.counter = self.counter_obj

        self.populate()
        self.block_ctl_obj.pull()
        self.block_list.push()
        self.page_list.push()

        self.scene_idx = 0

        info.push()

    def get_scene(self):
        self.scene_idx = (self.scene_idx + 1) % self.scene_max
        return self.scene_list._addr + self.scene_idx * 4

    def increment(self):
        self.counter_obj.count += 1
        self.counter_obj.push()

    def populate(self):
        idx = self.block_ctl.wptr.val
        total = self.block_ctl.total.val
        while idx < total:
            block = self.ctx.uobj.new_buf(self.block_size, "BM Block", track=False)
            self.block_list[idx * 2] = block._addr // self.page_size

            page_idx = idx * self.pages_per_block
            for i in range(self.pages_per_block):
                self.page_list[page_idx + i] = block._addr // self.page_size + i

            idx += 1
        self.block_ctl.wptr.val = idx




# SPDX-License-Identifier: MIT

import struct

from enum import IntEnum
from ..utils import *
from ..malloc import Heap

__all__ = ["DART8020Regs", "DART8020"]

class R_ERROR(Register32):
    FLAG = 31
    STREAM = 27, 24
    CODE = 23, 0
    NO_DAPF_MATCH = 11
    WRITE = 10
    SUBPAGE_PROT = 7
    PTE_READ_FAULT = 6
    READ_FAULT = 4
    WRITE_FAULT = 3
    NO_PTE = 2
    NO_PMD = 1
    NO_TTBR = 0

class R_STREAM_COMMAND(Register32):
    INVALIDATE = 20
    BUSY = 2

class R_TCR(Register32):
    BYPASS_DAPF = 12
    BYPASS_DART = 8
    TRANSLATE_ENABLE = 7

class R_TTBR(Register32):
    VALID = 31
    ADDR = 30, 0

class R_REMAP(Register32):
    MAP3 = 31, 24
    MAP2 = 23, 16
    MAP1 = 15, 8
    MAP0 = 7, 0

class PTE_T8020(Register64):
    SP_START = 63, 52
    SP_END = 51, 40
    OFFSET = 39, 14
    SP_PROT_DIS = 1
    VALID = 0

class PTE_T6000(Register64):
    SP_START = 63, 52
    SP_END = 51, 40
    OFFSET = 39, 10
    SP_PROT_DIS = 1
    VALID = 0

class R_CONFIG(Register32):
    LOCK = 15

class R_DAPF_LOCK(Register32):
    LOCK = 0

class DART8020Regs(RegMap):
    STREAM_COMMAND  = 0x20, R_STREAM_COMMAND
    STREAM_SELECT   = 0x34, Register32
    ERROR           = 0x40, R_ERROR
    ERROR_ADDR_LO   = 0x50, Register32
    ERROR_ADDR_HI   = 0x54, Register32
    CONFIG          = 0x60, R_CONFIG
    REMAP           = irange(0x80, 4, 4), R_REMAP

    DAPF_LOCK       = 0xf0, R_DAPF_LOCK
    UNK1            = 0xf8, Register32
    ENABLED_STREAMS = 0xfc, Register32

    TCR             = irange(0x100, 16, 4), R_TCR
    TTBR            = (irange(0x200, 16, 16), range(0, 16, 4)), R_TTBR

PTE_TYPES = {
    "dart,t8020": PTE_T8020,
    "dart,t6000": PTE_T6000,
}

class DART8020(Reloadable):
    PAGE_BITS = 14
    PAGE_SIZE = 1 << PAGE_BITS

    L0_SIZE = 4 # TTBR count
    L0_OFF = 36
    L1_OFF = 25
    L2_OFF = 14

    IDX_BITS = 11
    Lx_SIZE = (1 << IDX_BITS)
    IDX_MASK = Lx_SIZE - 1

    def __init__(self, iface, regs, util=None, compat="dart,t8020"):
        self.iface = iface
        self.regs = regs
        self.u = util
        self.pt_cache = {}
        self.enabled_streams = regs.ENABLED_STREAMS.val
        self.ptecls = PTE_TYPES[compat]

    @classmethod
    def from_adt(cls, u, path, instance=0, **kwargs):
        dart_addr = u.adt[path].get_reg(instance)[0]
        dart = cls(u.iface, dart_addr, u)
        dart.ptecls = PTE_TYPES[u.adt[path].compatible[0]]
        return dart

    def iomap_at(self, stream, iova, addr, size):
        if size == 0:
            return

        if not (self.enabled_streams & (1 << stream)):
            self.enabled_streams |= (1 << stream)
            self.regs.ENABLED_STREAMS.val |= self.enabled_streams

        tcr = self.regs.TCR[stream].reg

        if tcr.BYPASS_DART and not tcr.TRANSLATE_ENABLE:
            raise Exception("Stream is bypassed in DART")

        if tcr.BYPASS_DART or not tcr.TRANSLATE_ENABLE:
            raise Exception(f"Unknown DART mode {tcr}")

        if addr & (self.PAGE_SIZE - 1):
            raise Exception(f"Unaligned PA {addr:#x}")

        if iova & (self.PAGE_SIZE - 1):
            raise Exception(f"Unaligned IOVA {iova:#x}")

        start_page = align_down(iova, self.PAGE_SIZE)
        end = iova + size
        end_page = align_up(end, self.PAGE_SIZE)

        dirty = set()

        for page in range(start_page, end_page, self.PAGE_SIZE):
            paddr = addr + page - start_page

            l0 = page >> self.L0_OFF
            assert l0 < self.L0_SIZE
            ttbr = self.regs.TTBR[stream, l0].reg
            if not ttbr.VALID:
                l1addr = self.u.memalign(self.PAGE_SIZE, self.PAGE_SIZE)
                self.pt_cache[l1addr] = [0] * self.Lx_SIZE
                ttbr.VALID = 1
                ttbr.ADDR = l1addr >> 12
                self.regs.TTBR[stream, l0].reg = ttbr

            cached, l1 = self.get_pt(ttbr.ADDR << 12)
            l1idx = (page >> self.L1_OFF) & self.IDX_MASK
            l1pte = self.ptecls(l1[l1idx])
            if not l1pte.VALID:
                l2addr = self.u.memalign(self.PAGE_SIZE, self.PAGE_SIZE)
                self.pt_cache[l2addr] = [0] * self.Lx_SIZE
                l1pte = self.ptecls(
                    OFFSET=l2addr >> self.PAGE_BITS, VALID=1, SP_PROT_DIS=1)
                l1[l1idx] = l1pte.value
                dirty.add(ttbr.ADDR << 12)
            else:
                l2addr = l1pte.OFFSET << self.PAGE_BITS

            dirty.add(l1pte.OFFSET << self.PAGE_BITS)
            cached, l2 = self.get_pt(l2addr)
            l2idx = (page >> self.L2_OFF) & self.IDX_MASK
            self.pt_cache[l2addr][l2idx] = self.ptecls(
                SP_START=0, SP_END=0xfff,
                OFFSET=paddr >> self.PAGE_BITS, VALID=1, SP_PROT_DIS=1).value

        for page in dirty:
            self.flush_pt(page)

    def iotranslate(self, stream, start, size):
        if size == 0:
            return []

        tcr = self.regs.TCR[stream].reg

        if tcr.BYPASS_DART and not tcr.TRANSLATE_ENABLE:
            return [(start, size)]

        if tcr.BYPASS_DART or not tcr.TRANSLATE_ENABLE:
            raise Exception(f"Unknown DART mode {tcr}")

        start = start & 0xffffffff

        start_page = align_down(start, self.PAGE_SIZE)
        start_off = start - start_page
        end = start + size
        end_page = align_up(end, self.PAGE_SIZE)
        end_size = end - (end_page - self.PAGE_SIZE)

        pages = []

        for page in range(start_page, end_page, self.PAGE_SIZE):
            l0 = page >> self.L0_OFF
            assert l0 < self.L0_SIZE
            ttbr = self.regs.TTBR[stream, l0].reg
            if not ttbr.VALID:
                pages.append(None)
                continue

            cached, l1 = self.get_pt(ttbr.ADDR << 12)
            l1pte = self.ptecls(l1[(page >> self.L1_OFF) & self.IDX_MASK])
            if not l1pte.VALID and cached:
                cached, l1 = self.get_pt(ttbr.ADDR << 12, uncached=True)
                l1pte = self.ptecls(l1[(page >> self.L1_OFF) & self.IDX_MASK])
            if not l1pte.VALID:
                pages.append(None)
                continue

            cached, l2 = self.get_pt(l1pte.OFFSET << self.PAGE_BITS)
            l2pte = self.ptecls(l2[(page >> self.L2_OFF) & self.IDX_MASK])
            if not l2pte.VALID and cached:
                cached, l2 = self.get_pt(l1pte.OFFSET << self.PAGE_BITS, uncached=True)
                l2pte = self.ptecls(l2[(page >> self.L2_OFF) & self.IDX_MASK])
            if not l2pte.VALID:
                pages.append(None)
                continue

            pages.append(l2pte.OFFSET << self.PAGE_BITS)

        ranges = []

        for page in pages:
            if not ranges:
                ranges.append((page, self.PAGE_SIZE))
                continue
            laddr, lsize = ranges[-1]
            if ((page is None and laddr is None) or
                (page is not None and laddr == (page - lsize))):
                ranges[-1] = laddr, lsize + self.PAGE_SIZE
            else:
                ranges.append((page, self.PAGE_SIZE))

        ranges[-1] = (ranges[-1][0], ranges[-1][1] - self.PAGE_SIZE + end_size)

        if start_off:
            ranges[0] = (ranges[0][0] + start_off if ranges[0][0] else None,
                         ranges[0][1] - start_off)

        return ranges

    def get_pt(self, addr, uncached=False):
        cached = True
        if addr not in self.pt_cache or uncached:
            cached = False
            self.pt_cache[addr] = list(
                struct.unpack(f"<{self.Lx_SIZE}Q", self.iface.readmem(addr, self.PAGE_SIZE)))

        return cached, self.pt_cache[addr]

    def flush_pt(self, addr):
        assert addr in self.pt_cache
        self.iface.writemem(addr, struct.pack(f"<{self.Lx_SIZE}Q", *self.pt_cache[addr]))

    def initialize(self):
        for i in range(15):
            self.regs.TCR[i].reg = R_TCR(TRANSLATE_ENABLE=1)
        self.regs.TCR[15].reg = R_TCR(BYPASS_DART=1)

        for i in range(16):
            for j in range(4):
                self.regs.TTBR[i, j].reg = R_TTBR(VALID = 0)

        self.regs.ERROR.val = 0xffffffff
        self.regs.UNK1.val = 0
        self.regs.ENABLED_STREAMS.val = 0
        self.enabled_streams = 0

        self.invalidate_streams()

    def show_error(self):
        if self.regs.ERROR.reg.FLAG:
            print(f"ERROR: {self.regs.ERROR.reg!s}")
            print(f"ADDR: {self.regs.ERROR_ADDR_HI.val:#x}:{self.regs.ERROR_ADDR_LO.val:#x}")
            self.regs.ERROR.val = 0xffffffff

    def invalidate_streams(self, streams=0xffffffff):
        self.regs.STREAM_SELECT.val = streams
        self.regs.STREAM_COMMAND.val = R_STREAM_COMMAND(INVALIDATE=1)
        while self.regs.STREAM_COMMAND.reg.BUSY:
            pass

    def invalidate_cache(self):
        self.pt_cache = {}

    def dump_table2(self, base, l1_addr):

        def print_block(base, pte, start, last):
            pgcount = last - start
            pte.OFFSET -= pgcount
            print("    page (%4d): %08x ... %08x -> %016x [%d%d]" % (
                    start, base + start*0x4000, base + (start+1)*0x4000,
                    pte.OFFSET << self.PAGE_BITS, pte.SP_PROT_DIS, pte.VALID))
            if start < last:
                print("     ==> (%4d):          ... %08x -> %016x size: %08x" % (
                    last, base + (last+1)*0x4000,
                    (pte.OFFSET + pgcount - 1) << self.PAGE_BITS, pgcount << self.PAGE_BITS))

        cached, tbl = self.get_pt(l1_addr)

        unmapped = False
        start = 0
        next_pte = self.ptecls(VALID=0)

        for i, pte in enumerate(tbl):
            pte = self.ptecls(pte)
            if not pte.VALID:
                if not unmapped:
                    if next_pte.VALID:
                        print_block(base, next_pte, start, i)
                    print("  ...")
                    unmapped = True
                    next_pte = pte
                continue

            unmapped = False

            if int(pte) != int(next_pte):
                if next_pte.VALID:
                    print_block(base, next_pte, start, i)
                start = i

            next_pte = pte
            next_pte.OFFSET += 1

        if next_pte.VALID:
            print_block(base, next_pte, start, 2048)

    def dump_table(self, base, l1_addr):
        cached, tbl = self.get_pt(l1_addr)

        unmapped = False
        for i, pte in enumerate(tbl):
            pte = self.ptecls(pte)
            if not pte.VALID:
                if not unmapped:
                    print("  ...")
                    unmapped = True
                continue

            unmapped = False

            print("  table (%d): %08x ... %08x -> %016x [%d%d]" % (
                i, base + i*0x2000000, base + (i+1)*0x2000000,
                pte.OFFSET << self.PAGE_BITS, pte.SP_PROT_DIS, pte.VALID))
            self.dump_table2(base + i*0x2000000, pte.OFFSET << self.PAGE_BITS)

    def dump_ttbr(self, idx, ttbr):
        if not ttbr.VALID:
            return

        l1_addr = (ttbr.ADDR) << 12
        print("  TTBR%d: %09x" % (idx, l1_addr))

        self.dump_table(0, l1_addr)

    def dump_device(self, idx):
        tcr = self.regs.TCR[idx].reg
        ttbrs = self.regs.TTBR[idx, :]
        print(f"dev {idx:02x}: TCR={tcr!s} TTBRs = [{', '.join(map(str, ttbrs))}]")

        if tcr.TRANSLATE_ENABLE and tcr.BYPASS_DART:
            print("  mode: INVALID")
        elif tcr.TRANSLATE_ENABLE:
            print("  mode: TRANSLATE")

            for idx, ttbr in enumerate(ttbrs):
                self.dump_ttbr(idx, ttbr.reg)
        elif tcr.BYPASS_DART:
            print("  mode: BYPASS")
        else:
            print("  mode: UNKNOWN")

    def dump_params(self):
        pass



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib, time
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *

from m1n1.fw.isp_base import ISP
from m1n1.fw.isp.isp_vid import ISPFrameReceiver


isp = ISP(u)
isp.boot()

rcver = ISPFrameReceiver(isp)
rcver.stream()  # Press 'Enter' to exit stream



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *
from m1n1 import asm

ULCON = 0x235200000
UCON = 0x235200004
UFCON = 0x235200008
UTRSTAT = 0x235200010

AIC = 0x23b100000

AIC_RST = AIC + 0xc
AIC_CFG = AIC + 0x10

AIC_TB = 0x23b108000
AIC_TGT_DST = AIC + 0x3000
AIC_SW_GEN_SET = AIC + 0x4000
AIC_SW_GEN_CLR = AIC + 0x4080
AIC_MASK_SET = AIC + 0x4100
AIC_MASK_CLR = AIC + 0x4180
AIC_HW_STATE = AIC + 0x4200

AIC_INTERRUPT_ACK = AIC + 0x2004
AIC_IPI_SET = AIC + 0x2008
AIC_IPI_CLR = AIC + 0x200c

AIC_IPI_MASK_SET = AIC + 0x2024
AIC_IPI_MASK_CLR = AIC + 0x2028

daif = u.mrs(DAIF)
print("DAIF: %x" % daif)
daif &= ~0x3c0
#daif |= 0x3c0
u.msr(DAIF, daif)
print("DAIF: %x" % u.mrs(DAIF))

def cpoll():
    mon.poll()
    print("<")
    mon.poll()
    print(">")

p.write32(AIC + 0xc, 1)
p.write32(AIC + 0x10, 0xe0777971)
p.write32(AIC + 0x18, 0)
p.write32(AIC + 0x20, 0xffffffff)
p.write32(AIC + 0x24, 0xffffffff)
p.write32(AIC + 0x28, 0xffffffff)
p.write32(AIC + 0x2c, 0xffffffff)
p.write32(AIC + 0x30, 0xffffffff)
p.write32(AIC + 0x34, 0xffffffff)
p.write32(AIC + 0x38, 0xffffffff)
p.write32(AIC + 0x3c, 0xffffffff)
p.write32(AIC + 0x40, 0xffffffff)
p.write32(AIC + 0x38, 0xffffffff)
#p.write32(AIC + 0xc, 0)

p.memset32(AIC_MASK_SET, 0xffffffff, 0x80)
p.memset32(AIC_SW_GEN_CLR, 0xffffffff, 0x80)
p.memset32(AIC_TGT_DST, 0x1, 0x1000)
#p.memset32(AIC_MASK_CLR, 0xffffffff, 0x80)

#p.write32(AIC + 0x10, 0xe0777971)

mon.add(AIC + 0x0000, 0x1000)
mon.add(AIC + 0x2080, 0x040)
mon.add(AIC + 0x4000, 0x200)
mon.add(AIC + 0x5000, 0x080)
mon.add(AIC + 0x5080, 0x080)
mon.add(AIC + 0x5100, 0x080)
mon.add(AIC + 0x5180, 0x080)
mon.add(AIC + 0x5200, 0x080)
mon.add(AIC + 0x5280, 0x080)
mon.add(AIC + 0x5300, 0x080)
mon.add(AIC + 0x5380, 0x080)
#mon.add(AIC + 0x3000, 0x400)
#mon.add(AIC + 0x4000, 0x400)
#mon.add(AIC + 0x8000, 0x20)
#mon.add(AIC + 0x8030, 0xd0)
#mon.add(0x235200000, 0x20)

def test_ipi():
    cpoll()

    print("Set IPI")

    p.write32(AIC_IPI_SET, 1)

    cpoll()
    cpoll()

    print("Read ACK reg")

    reason = p.read32(AIC_INTERRUPT_ACK)
    print("reason: 0x%x" % reason)

    cpoll()

    print("Write reason")
    p.write32(AIC_INTERRUPT_ACK, reason)

    cpoll()

    reason = p.read32(AIC_INTERRUPT_ACK)
    print("reason: 0x%x" % reason)

    cpoll()

    print("Write ACK reg")
    p.write32(AIC_INTERRUPT_ACK, reason)
    cpoll()

    print("Clear IPI")

    p.write32(AIC_IPI_CLR, 1)
    cpoll()

    print("Read ACK reg")

    reason = p.read32(AIC_INTERRUPT_ACK)

    print("reason: 0x%x" % reason)

    cpoll()

    print("Write IPI ACK")

    p.write32(AIC_IPI_MASK_CLR, 1)

    cpoll()

def test_timer():
    cpoll()

    freq = u.mrs(CNTFRQ_EL0)
    print("Timer freq: %d" % freq)

    #u.msr(CNTP_CTL_EL0, 0)
    #u.msr(CNTP_TVAL_EL0, freq * 2)
    #u.msr(CNTP_CTL_EL0, 1)
    #u.msr(CNTV_CTL_EL0, 0)
    #u.msr(CNTV_TVAL_EL0, freq * 2)
    #u.msr(CNTV_CTL_EL0, 1)
    #u.msr(CNTHV_CTL_EL2, 0)
    #u.msr(CNTHV_TVAL_EL2, freq * 2)
    #u.msr(CNTHV_CTL_EL2, 1)
    u.msr(CNTHP_CTL_EL2, 0)
    u.msr(CNTHP_TVAL_EL2, freq * 2)
    u.msr(CNTHP_CTL_EL2, 1)

    iface.ttymode()

    #while True:
        #p.nop()
        #time.sleep(0.3)
        #print(". %x" % u.mrs(CNTP_CTL_EL0))

def get_irq_state(irq):
    v = p.read32(AIC_HW_STATE + 4* (irq//32))
    return bool(v & 1<<(irq%32))

def test_uart_irq():
    cpoll()
    #p.memset32(AIC_MASK_CLR, 0xffffffff, 0x80)

    print("cleanup")
    p.write32(UCON, 5)
    p.write32(UFCON, 0x11)
    p.write32(UTRSTAT, 0xfff)

    cpoll()

    for irq in range(600, 610):
        #print("S: ", get_irq_state(irq))
        p.write32(AIC_SW_GEN_CLR + 4* (irq//32), 1<<(irq%32))
        #print("S: ", get_irq_state(irq))
        #print("a")
        #print("S: ", get_irq_state(irq))
        p.write32(AIC_MASK_CLR + 4* (irq//32), 1<<(irq%32))
        #print("S: ", get_irq_state(irq))
        #print("b")

    irq = 605

    cpoll()
    print("a")
    print("S: ", get_irq_state(irq))
    print("ucon: %x" %p.read32(UCON))

    TX_IRQ_EN = 0x1000

    RX_IRQ_ENABLE = 0x20000
    RX_IRQ_UNMASK = 0x10000

    RX_IRQ_ENA = 0x20000
    RX_IRQ_MASK = 0x4000 # defer?

    code = u.malloc(0x1000)

    c = asm.ARMAsm("""
        ldr x1, =0x235200000

        ldr x3, =0xc000000
1:
        subs x3, x3, #1
        bne 1b
        mov x2, 'A'
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]
        #str w2, [x1, #0x20]

        mov x3, #0x3ff
        str w3, [x1, #0x10]
        #str w2, [x1, #0x20]
        str w0, [x1, #4]
        ldr w0, [x1, #0x10]

        ldr x3, =0xc00000
1:
        subs x3, x3, #1
        bne 1b

        #mov x3, #0x3ff
        #str w3, [x1, #0x10]
        #ldr w2, [x1, #4]
        #mov x2, #0x205
        #str w2, [x1, #4]
        #str w0, [x1, #4]
        ##ldr w0, [x1, #0x10]

        #ldr x3, =0xc00000
#1:
        #subs x3, x3, #1
        #bne 1b

        ldr w0, [x1, #0x10]
        #mov w0, w2
        ret
""", code)
    iface.writemem(code, c.data)
    p.dc_cvau(code, len(c.data))
    p.ic_ivau(code, len(c.data))

    #RX_IRQ_

    """
UCON    UTRSTAT
00200           TX FIFO thresh IRQ delivery enable
00080   0200    TX FIFO threshold IRQ unmask
20000   0100    RX IRQ unmask
10000           RX IRQ delivery enable
"""

    # edge triggered
    TX_FIFO_THRESH_CROSSED_IRQ_UNMASK = 0x2000

    TX_IRQ_UNMASK = 0x200
    TX_EVENT_ENABLE = 0x80
    RX_EVENT_ENABLE = 0x20000
    RX_IRQ_UNMASK = 0x10000

    #flags = 0x7ffc0
    crash = 0x180000
    no_irqs = 0x21c5c0
    instant_irqs = 0x3a00
    #flags = no_irqs | 0x0000
    #flags = 0x2e5c0
    #flags = 0x2000

    #flags = 0x30000
    #flags = 0x80
    flags = 0x7ff80


    val = flags | 0x005
    #print("ucon<-%x" % val)
    #p.write32(UCON, val)
    p.write32(UTRSTAT, 0xfff)
    print("utrstat=%x" % p.read32(UTRSTAT))
    ret = p.call(code, val)
    print("utrstat::%x" % ret)
    print("utrstat=%x" % p.read32(UTRSTAT))
    time.sleep(0.5)
    iface.dev.write(b'1')
    #print(iface.dev.read(1))
    time.sleep(0.1)
    print("ucon: %x" %p.read32(UCON))
    print("delay")
    try:
        p.udelay(500000)
    except:
        pass
    iface.dev.write(bytes(64))
    p.nop()
    print("ucon: %x" %p.read32(UCON))
    print("S: ", get_irq_state(irq))

    #while True:
        #print("S: ", get_irq_state(irq))
        #p.write32(UTRSTAT, 0xfff)
        #print("utrstat=%x" % p.read32(UTRSTAT))
        #print("ucon: %x" %p.read32(UCON))
        #print(">S: ", get_irq_state(irq))
        #p.write32(UCON, flags | 0x005)
        #print(">ucon: %x" %p.read32(UCON))
        #time.sleep(0.1)



def test_smp_ipi():
    p.smp_start_secondaries()

    code = u.malloc(0x1000)

    c = asm.ARMAsm("""
#define sys_reg(op0, op1, CRn, CRm, op2) s##op0##_##op1##_c##CRn##_c##CRm##_##op2
#define SYS_CYC_OVRD           sys_reg(3, 5, 15, 5, 0)

        msr DAIFClr, 7
        ldr x1, =0x000000
        msr SYS_CYC_OVRD, x1
        mrs x0, SYS_CYC_OVRD
        mov x1, #0x1000000
1:
        subs x1, x1, #1
        mrs x0, HCR_EL2
        bne 1b
        ret
""", code)

    iface.writemem(code, c.data)
    p.dc_cvau(code, len(c.data))
    p.ic_ivau(code, len(c.data))

    print("Enable IRQs on secondaries")
    for i in range(1, 8):
        ret = p.smp_call_sync(i, code)
        print("0x%x"%ret)

    #e0477971
    #p.write32(AIC + 0x10, 0xe0777971)
    #p.write32(AIC + 0x28, 0xffffffff)

    cpoll()

    print("Clear IPI")
    p.write32(AIC_IPI_CLR, 0xffffffff)
    p.write32(AIC_IPI_MASK_CLR, 0xffffffff)
    for i in range(8):
        p.write32(AIC_IPI_CLR+0x3000+i*0x80, 0xffffffff)
        p.write32(AIC_IPI_MASK_CLR+0x3000+i*0x80, 0xffffffff)

    cpoll()

    print("Set IPI")
    #p.write32(AIC_IPI_SET, 0x00000004)
    #p.write32(AIC_IPI_SET, 0x00000000)

    cpoll()
    print("Clear IPI")
    p.write32(AIC_IPI_CLR, 0xffffffff)
    p.write32(AIC_IPI_MASK_CLR, 0xffffffff)
    for i in range(8):
        p.write32(AIC_IPI_CLR+0x3000+i*0x80, 1)
        p.write32(AIC_IPI_MASK_CLR+0x3000+i*0x80, 1)

def test_smp_affinity():
    p.write32(AIC_TGT_DST, 0x6)
    p.write32(AIC_TGT_DST+4, 0xfe)
    p.write32(AIC_TGT_DST+8, 0xfe)
    p.write32(AIC_TGT_DST+12, 0x6)
    p.write32(AIC_SW_GEN_SET,0x8);
    p.write32(AIC_MASK_CLR,0x8);

#test_ipi()
#test_timer()
#test_uart_irq()
test_smp_ipi()
test_smp_affinity()



# SPDX-License-Identifier: MIT

import struct

from enum import IntEnum

from m1n1.proxyutils import RegMonitor
from m1n1.utils import *
from m1n1.trace.dart import DARTTracer
from m1n1.trace.asc import ASCTracer, EP, EPState, msg, msg_log, DIR
from m1n1.fw.smc import *

ASCTracer = ASCTracer._reloadcls()

class SMCEpTracer(EP):
    BASE_MESSAGE = SMCMessage

    def __init__(self, tracer, epid):
        super().__init__(tracer, epid)
        self.state.sram_addr = None
        self.state.verbose = 1
        self.state.rb = {}

    Initialize = msg_log(SMC_INITIALIZE, DIR.TX, SMCInitialize)
    Notification = msg_log(SMC_NOTIFICATION, DIR.RX)

    @msg(SMC_WRITE_KEY, DIR.TX, SMCWriteKey)
    def WriteKey(self, msg):
        key = msg.KEY.to_bytes(4, byteorder="big").decode("ascii")
        self.state.rb[msg.ID] = msg.TYPE, key, msg.SIZE
        data = self.hv.iface.readmem(self.state.sram_addr, msg.SIZE)
        self.log(f"[{msg.ID:x}] >W: <{key}> = {data.hex()} ({msg.SIZE})")
        return True

    @msg(SMC_READ_KEY, DIR.TX, SMCReadKey)
    def ReadKey(self, msg):
        key = msg.KEY.to_bytes(4, byteorder="big").decode("ascii")
        self.state.rb[msg.ID] = msg.TYPE, key, msg.SIZE
        self.log(f"[{msg.ID:x}] >R: <{key}> = ... ({msg.SIZE})")
        return True

    @msg(SMC_RW_KEY, DIR.TX, SMCReadWriteKey)
    def ReadKeyPayload(self, msg):
        key = msg.KEY.to_bytes(4, byteorder="big").decode("ascii")
        self.state.rb[msg.ID] = msg.TYPE, key, msg.RSIZE
        data = self.hv.iface.readmem(self.state.sram_addr, msg.WSIZE)
        self.log(f"[{msg.ID:x}] >RP: <{key}> = {data.hex()} ({msg.WSIZE, msg.RSIZE})")
        return True

    @msg(SMC_GET_KEY_INFO, DIR.TX, SMCGetKeyInfo)
    def GetInfo(self, msg):
        key = msg.KEY.to_bytes(4, byteorder="big").decode("ascii")
        self.state.rb[msg.ID] = msg.TYPE, key, None
        self.log(f"[{msg.ID:x}] >KInfo: <{key}>")
        return True

    @msg(SMC_GET_KEY_BY_INDEX, DIR.TX, SMCGetKeyByIndex)
    def GetKeyByIndex(self, msg):
        self.state.rb[msg.ID] = msg.TYPE, msg.INDEX, None
        self.log(f"[{msg.ID:x}] >KIdx: <{msg.INDEX}>")
        return True

    @msg(None, DIR.RX, Register64)
    def RXMsg(self, msg):
        if self.state.sram_addr is None:
            self.log(f"SRAM address: {msg.value:#x}")
            self.state.sram_addr = msg.value
            return True

        msg = SMCResult(msg.value)
        if msg.RESULT != 0:
            self.log(f"[{msg.ID:x}] <Err: 0x{msg.RESULT:02x}")
            return True

        if msg.ID in self.state.rb:
            msgtype, key, size = self.state.rb.pop(msg.ID)
            if msgtype in (SMC_READ_KEY, SMC_RW_KEY):
                if size <= 4:
                    data = hex(msg.VALUE)
                else:
                    data = self.hv.iface.readmem(self.state.sram_addr, msg.SIZE).hex()
                self.log(f"[{msg.ID:x}] <R: <{key}> = {data}")
                return True

            elif msgtype == SMC_GET_KEY_INFO:
                data = self.hv.iface.readmem(self.state.sram_addr, 6)
                size, type, flags = struct.unpack("B4sB", data)
                self.log(f"[{msg.ID:x}] <Info: <{key}>: size={size} type={type.decode('ascii')} flags={flags:#x}")
                return True

            elif msgtype == SMC_GET_KEY_BY_INDEX:
                kname = msg.VALUE.to_bytes(4, byteorder="little").decode("ascii")
                self.log(f"[{msg.ID:x}] <Key @{key}: <{kname}>")
                return True

        self.log(f"[{msg.ID:x}] <OK {msg!r}")
        return True

class SMCTracer(ASCTracer):
    ENDPOINTS = {
        0x20: SMCEpTracer
    }

    def handle_msg(self, direction, r0, r1):
        super().handle_msg(direction, r0, r1)

smc_tracer = SMCTracer(hv, "/arm-io/smc", verbose=1)
smc_tracer.start()



# SPDX-License-Identifier: MIT
from construct import *

__all__ = ["BootArgs_r1", "BootArgs_r2", "BootArgs_r3"]

BootArgs_r1 = Struct(
    "revision"              / Hex(Int16ul),
    "version"               / Hex(Int16ul),
    Padding(4),
    "virt_base"             / Hex(Int64ul),
    "phys_base"             / Hex(Int64ul),
    "mem_size"              / Hex(Int64ul),
    "top_of_kernel_data"    / Hex(Int64ul),
    "video" / Struct(
        "base"      / Hex(Int64ul),
        "display"   / Hex(Int64ul),
        "stride"    / Hex(Int64ul),
        "width"     / Hex(Int64ul),
        "height"    / Hex(Int64ul),
        "depth"     / Hex(Int64ul),
    ),
    "machine_type"          / Hex(Int32ul),
    Padding(4),
    "devtree"               / Hex(Int64ul),
    "devtree_size"          / Hex(Int32ul),
    "cmdline"               / PaddedString(256, "ascii"),
    Padding(4),
    "boot_flags"            / Hex(Int64ul),
    "mem_size_actual"       / Hex(Int64ul),
)

BootArgs_r2 = Struct(
    "revision"              / Hex(Int16ul),
    "version"               / Hex(Int16ul),
    Padding(4),
    "virt_base"             / Hex(Int64ul),
    "phys_base"             / Hex(Int64ul),
    "mem_size"              / Hex(Int64ul),
    "top_of_kernel_data"    / Hex(Int64ul),
    "video" / Struct(
        "base"      / Hex(Int64ul),
        "display"   / Hex(Int64ul),
        "stride"    / Hex(Int64ul),
        "width"     / Hex(Int64ul),
        "height"    / Hex(Int64ul),
        "depth"     / Hex(Int64ul),
    ),
    "machine_type"          / Hex(Int32ul),
    Padding(4),
    "devtree"               / Hex(Int64ul),
    "devtree_size"          / Hex(Int32ul),
    "cmdline"               / PaddedString(608, "ascii"),
    Padding(4),
    "boot_flags"            / Hex(Int64ul),
    "mem_size_actual"       / Hex(Int64ul),
)

BootArgs_r3 = Struct(
    "revision"              / Hex(Int16ul),
    "version"               / Hex(Int16ul),
    Padding(4),
    "virt_base"             / Hex(Int64ul),
    "phys_base"             / Hex(Int64ul),
    "mem_size"              / Hex(Int64ul),
    "top_of_kernel_data"    / Hex(Int64ul),
    "video" / Struct(
        "base"      / Hex(Int64ul),
        "display"   / Hex(Int64ul),
        "stride"    / Hex(Int64ul),
        "width"     / Hex(Int64ul),
        "height"    / Hex(Int64ul),
        "depth"     / Hex(Int64ul),
    ),
    "machine_type"          / Hex(Int32ul),
    Padding(4),
    "devtree"               / Hex(Int64ul),
    "devtree_size"          / Hex(Int32ul),
    "cmdline"               / PaddedString(1024, "ascii"),
    Padding(4),
    "boot_flags"            / Hex(Int64ul),
    "mem_size_actual"       / Hex(Int64ul),
)



# SPDX-License-Identifier: MIT

import struct

from enum import IntEnum
from ..utils import *
from ..malloc import Heap

from .dart8020 import DART8020, DART8020Regs
from .dart8110 import DART8110, DART8110Regs

__all__ = ["DART"]

class DART(Reloadable):
    PAGE_BITS = 14
    PAGE_SIZE = 1 << PAGE_BITS

    def __init__(self, iface, regs, util=None, compat="dart,t8020", iova_range=(0x80000000, 0x90000000)):
        self.iface = iface
        self.iova_allocator = [Heap(iova_range[0], iova_range[1], self.PAGE_SIZE)
                               for i in range(16)]
        if compat in ["dart,t8020", "dart,t6000"]:
            self.dart = DART8020(iface, regs, util, compat)
        elif compat in ["dart,t8110"]:
            self.dart = DART8110(iface, regs, util)
        else:
            raise TypeError(compat)

    @classmethod
    def from_adt(cls, u, path, instance=0, **kwargs):
        dart_addr = u.adt[path].get_reg(instance)[0]
        compat = u.adt[path].compatible[0]
        if compat in ["dart,t8020", "dart,t6000"]:
            regs = DART8020Regs(u, dart_addr)
        elif compat in ["dart,t8110"]:
            regs = DART8110Regs(u, dart_addr)
        return cls(u.iface, regs, u, compat, **kwargs)

    def ioread(self, stream, base, size):
        if size == 0:
            return b""

        ranges = self.iotranslate(stream, base, size)

        iova = base
        data = []
        for addr, size in ranges:
            if addr is None:
                raise Exception(f"Unmapped page at iova {iova:#x}")
            data.append(self.iface.readmem(addr, size))
            iova += size

        return b"".join(data)

    def iowrite(self, stream, base, data):
        if len(data) == 0:
            return

        ranges = self.iotranslate(stream, base, len(data))

        iova = base
        p = 0
        for addr, size in ranges:
            if addr is None:
                raise Exception(f"Unmapped page at iova {iova:#x}")
            self.iface.writemem(addr, data[p:p + size])
            p += size
            iova += size

    def iomap(self, stream, addr, size):
        iova = self.iova_allocator[stream].malloc(size)

        self.iomap_at(stream, iova, addr, size)
        return iova

    def iomap_at(self, stream, iova, addr, size):
        self.dart.iomap_at(stream, iova, addr, size)

    def iotranslate(self, stream, start, size):
        return self.dart.iotranslate(stream, start, size)

    def initialize(self):
        self.dart.initialize()

    def show_error(self):
        self.dart.show_error()

    def invalidate_streams(self, streams=0xffffffff):
        self.dart.invalidate_streams(streams)

    def invalidate_cache(self):
        self.dart.invalidate_cache()

    def dump_device(self, idx):
        self.dart.dump_device(idx)

    def dump_all(self):
        for i in range(16):
            self.dump_device(i)

    def dump_params(self):
        self.dart.dump_params()



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

import struct

from m1n1.setup import *
from m1n1 import asm

base = 0x235010000

# register defines from https://github.com/torvalds/linux/blob/master/drivers/i2c/busses/i2c-pasemi.c
# Copyright (C) 2006-2007 PA Semi, Inc
# SMBus host driver for PA Semi PWRficient
REG_MTXFIFO = 0x00
REG_MRXFIFO = 0x04
REG_SMSTA = 0x14
REG_CTL = 0x1c

MTXFIFO_READ = 0x00000400
MTXFIFO_STOP = 0x00000200
MTXFIFO_START = 0x00000100
MTXFIFO_DATA_M = 0x000000ff

MRXFIFO_EMPTY = 0x00000100
MRXFIFO_DATA_M = 0x000000ff

SMSTA_XEN = 0x08000000
SMSTA_MTN = 0x00200000

CTL_MRR = 0x00000400
CTL_MTR = 0x00000200
CTL_CLK_M = 0x000000ff

CLK_100K_DIV = 84
CLK_400K_DIV = 21


def i2c_read_reg(addr, reg, reg_size):
    p.set32(base + REG_CTL, CTL_MTR | CTL_MRR)
    p.write32(base + REG_SMSTA, 0xffffffff)

    p.write32(base + REG_MTXFIFO, MTXFIFO_START | (addr << 1))
    p.write32(base + REG_MTXFIFO, MTXFIFO_STOP | reg)

    while not (p.read32(base + REG_SMSTA) & SMSTA_XEN):
        pass

    p.write32(base + REG_MTXFIFO, MTXFIFO_START | (addr << 1) | 1)
    p.write32(base + REG_MTXFIFO, MTXFIFO_READ | MTXFIFO_STOP | reg_size + 1)

    res = []
    while len(res) < reg_size+1:
        v = p.read32(base + REG_MRXFIFO)
        if v & 0x100:
            continue
        res.append(v)

    if res[0] < reg_size:
        print("only read %d instead of %d bytes" % (res[0], reg_size))
    return res[1:]


def i2c_write_reg(addr, reg, data):
    p.set32(base + REG_CTL, CTL_MTR | CTL_MRR)
    p.write32(base + REG_SMSTA, 0xffffffff)

    p.write32(base + REG_MTXFIFO, MTXFIFO_START | (addr << 1))
    p.write32(base + REG_MTXFIFO, reg)
    for i in range(len(data)-1):
        p.write32(base + REG_MTXFIFO, data[i])
    p.write32(base + REG_MTXFIFO, data[-1] | MTXFIFO_STOP)

    while not (p.read32(base + REG_SMSTA) & SMSTA_XEN):
        pass


def i2c_read16(addr, reg):
    data = struct.pack(">2b", *i2c_read_reg(addr, reg, 2))
    return struct.unpack(">H", data)[0]


def i2c_read32(addr, reg):
    data = struct.pack(">4b", *i2c_read_reg(addr, reg, 4))
    return struct.unpack(">I", data)[0]


def tps6598x_exec_cmd(addr, cmd, data_in, out_len):
    if data_in:
        data = [len(data_in)] + data_in

        # TPS_REG_DATA1
        i2c_write_reg(addr, 0x09, data)

    # TPS_REG_CMD1
    cmd = [4] + list(map(ord, cmd))
    i2c_write_reg(addr, 0x08, cmd)

    # TPS_REG_CMD1
    v = i2c_read32(addr, 0x08)
    while v != 0:
        if v == 0x21434d44:  # !CMD
            raise Exception("Invalid command!")
        v = i2c_read32(addr, 0x08)

    if not out_len:
        return

    # TPS_REG_DATA1
    return i2c_read_reg(addr, 0x09, out_len)


print("make sure to run pmgr_adt_clocks_enable for /arm-io/i2c0 before this script.")

# apple-specific command to bring the power state to zero
# (or any other value specified as an argument)
tps6598x_exec_cmd(0x3f, "SSPS", [0], 0)
tps6598x_exec_cmd(0x38, "SSPS", [0], 0)

tps6598x_exec_cmd(0x3f, "SWDF", None, 0)
tps6598x_exec_cmd(0x3f, "SWSr", None, 0)
tps6598x_exec_cmd(0x38, "SWDF", None, 0)
tps6598x_exec_cmd(0x38, "SWSr", None, 0)



# SPDX-License-Identifier: MIT
import struct

from ..utils import *
from .spmi import SPMI

__all__ = ["PMU"]

class PMU:

    def __init__(self, u, adt_path=None):
        self.u = u
        if adt_path is None:
            adt_path = PMU.find_primary_pmu(u.adt)

        self.node = u.adt[adt_path]
        self.spmi = SPMI(u, adt_path.rpartition('/')[0])
        self.adt_path = adt_path
        self.primary = u.adt[adt_path].is_primary == 1
        self.reg = u.adt[adt_path].reg[0]

    def reset_panic_counter(self):
        if self.primary:
            leg_scrpad = self.node.info_leg__scrpad[0]
            self.spmi.write8(self.reg, leg_scrpad + 2, 0) # error counts

    @staticmethod
    def find_primary_pmu(adt):
        for child in adt["/arm-io"]:
            if child.name.startswith("nub-spmi"):
                for pmu in child:
                    compat = getattr(pmu, "compatible")[0] if hasattr(pmu, "compatible") else "unset"
                    primary = (getattr(pmu, "is-primary") == 1) if hasattr(pmu, "is-primary")  else False
                    if compat == "pmu,spmi" and primary:
                        return pmu._path.removeprefix('/device-tree')
        raise KeyError(f"primary 'pmu,spmi' node not found")



# SPDX-License-Identifier: MIT
import datetime

from m1n1.constructutils import show_struct_trace, Ver
from m1n1.utils import *

Ver.set_version(hv.u)

trace_device("/arm-io/sgx", False)
trace_device("/arm-io/pmp", False)
trace_device("/arm-io/gfx-asc", False)

from m1n1.trace.agx import AGXTracer
AGXTracer = AGXTracer._reloadcls(True)

agx_tracer = AGXTracer(hv, "/arm-io/gfx-asc", verbose=1)
agx_tracer.trace_usermap = False
agx_tracer.trace_kernmap = False
agx_tracer.trace_userva = False
agx_tracer.trace_kernva = False
agx_tracer.redump = False
agx_tracer.cmd_dump_dir = "gfxdumps"

#agx_tracer.encoder_id_filter = lambda i: (i >> 16) == 0xc0de
agx_tracer.start()

def resume_tracing(ctx):
    fname = f"{datetime.datetime.now().isoformat()}.log"
    hv.set_logfile(open(f"gfxlogs/{fname}", "a"))
    agx_tracer.resume()
    return True

def pause_tracing(ctx):
    agx_tracer.pause()
    hv.set_logfile(None)
    return True

hv.add_hvcall(100, resume_tracing)
hv.add_hvcall(101, pause_tracing)

mode = TraceMode.SYNC
trace_range(irange(agx_tracer.gpu_region, agx_tracer.gpu_region_size), mode=mode, name="gpu_region")
trace_range(irange(agx_tracer.gfx_shared_region, agx_tracer.gfx_shared_region_size), mode=mode, name="gfx_shared_region")

## Trace the entire mmio range around the GPU
node = hv.adt["/arm-io/sgx"]
addr, size = node.get_reg(0)
hv.trace_range(irange(addr, 0x1000000), TraceMode.SYNC, name="sgx")
#hv.trace_range(irange(addr, 0x1000000), TraceMode.OFF, name="sgx")
hv.trace_range(irange(0x204017030, 8), TraceMode.SYNC, name="faultcode")

trace_device("/arm-io/sgx", True)
trace_device("/arm-io/gfx-asc", False)

def trace_all_gfx_io():
    # These are all the IO ranges that get mapped into the UAT iommu pagetable
    # Trace them so we can see if any of them are being written by the CPU

    # page (8): fa010020000 ... fa010023fff -> 000000020e100000 [8000020e100447]
    hv.trace_range(irange(0x20e100000, 0x4000), mode=TraceMode.SYNC)

    # page (10): fa010028000 ... fa01002bfff -> 000000028e104000 [c000028e104447]
    hv.trace_range(irange(0x20e100000, 0x4000), mode=TraceMode.SYNC)

    # page (22): fa010058000 ... fa01005bfff -> 000000028e494000 [8000028e494447]
    hv.trace_range(irange(0x28e494000, 0x4000), mode=TraceMode.SYNC)

    # page (28): fa010070000 ... fa010073fff -> 0000000204d60000 [c0000204d60447]
    hv.trace_range(irange(0x204d60000, 0x4000), mode=TraceMode.SYNC)

    # page (30): fa010078000 ... fa01007bfff -> 0000000200000000 [c0000200000447]
    #    to
    # page (83): fa01014c000 ... fa01014ffff -> 00000002000d4000 [c00002000d4447]
    hv.trace_range(irange(0x200000000, 0xd5000), mode=TraceMode.SYNC)

    # page (84): fa010150000 ... fa010153fff -> 0000000201000000 [c0000201000447]
    #page (137): fa010224000 ... fa010227fff -> 00000002010d4000 [c00002010d4447]
    hv.trace_range(irange(0x201000000, 0xd5000), mode=TraceMode.SYNC)

    # page (138): fa010228000 ... fa01022bfff -> 0000000202000000 [c0000202000447]
    # page (191): fa0102fc000 ... fa0102fffff -> 00000002020d4000 [c00002020d4447]
    hv.trace_range(irange(0x202000000, 0xd5000), mode=TraceMode.SYNC)

    # page (192): fa010300000 ... fa010303fff -> 0000000203000000 [c0000203000447]
    hv.trace_range(irange(0x203000000, 0xd5000), mode=TraceMode.SYNC)
    hv.trace_range(irange(0x204000000, 0xd5000), mode=TraceMode.SYNC)
    hv.trace_range(irange(0x205000000, 0xd5000), mode=TraceMode.SYNC)
    hv.trace_range(irange(0x206000000, 0xd5000), mode=TraceMode.SYNC)
    hv.trace_range(irange(0x207000000, 0xd5000), mode=TraceMode.SYNC)

    # page (464): fa010740000 ... fa010743fff -> 00000002643c4000 [c00002643c4447]
    hv.trace_range(irange(0x2643c4000, 0x4000), mode=TraceMode.SYNC)
    # page (466): fa010748000 ... fa01074bfff -> 000000028e3d0000 [c000028e3d0447]
    hv.trace_range(irange(0x28e3d0000, 0x4000), mode=TraceMode.SYNC)
    # page (468): fa010750000 ... fa010753fff -> 000000028e3c0000 [8000028e3c0447]
    hv.trace_range(irange(0x28e3c0000, 0x4000), mode=TraceMode.SYNC)

    # page (8): f9100020000 ... f9100023fff -> 0000000406000000 [60000406000447]
    # page (263): f910041c000 ... f910041ffff -> 00000004063fc000 [600004063fc447]
    hv.trace_range(irange(0x2643c4000, 0x63fc000), mode=TraceMode.SYNC)

def trace_gpu_irqs():
    # Trace sgx interrupts
    node = hv.adt["/arm-io/sgx"]
    for irq in getattr(node, "interrupts"):
        hv.trace_irq(f"{node.name} {irq}", irq, 1, hv.IRQTRACE_IRQ)

    ## Trace gfx-asc interrupts
    #node = hv.adt["/arm-io/gfx-asc"]
    #for irq in getattr(node, "interrupts"):
        #hv.trace_irq(f"{node.name} {irq}", irq, 1, hv.IRQTRACE_IRQ)

trace_gpu_irqs()



# SPDX-License-Identifier: MIT
# Extension for ipython to handle monitor polling
# and simd context after each executed line


class PostExecuteWatcher(object):
    def __init__(self, ip):
        self.shell = ip

    def post_execute(self):
        mon = self.shell.user_ns.get("mon", None)
        if mon != None:
            try:
                mon.poll()
            except Exception as e:
                print(f"mon.poll() failed: {e!r}")
        u = self.shell.user_ns.get("u", None)
        if u != None:
            u.push_simd()


def load_ipython_extension(ip):
    pew = PostExecuteWatcher(ip)
    ip.events.register("post_execute", pew.post_execute)



# SPDX-License-Identifier: MIT
from ...utils import *

from ..asc import StandardASC
from .aopep import *
from .base import AOPBase

class AOPClient(StandardASC, AOPBase):
    ENDPOINTS = {
        0x20: AOPSPUAppEndpoint,
        0x21: AOPAccelEndpoint,
        0x22: AOPGyroEndpoint,
        0x24: AOPALSEndpoint,
        0x25: AOPWakehintEndpoint,
        0x26: AOPUNK26Endpoint,
        0x27: AOPAudioEndpoint,
        0x28: AOPVoiceTriggerEndpoint,
    }
    def __init__(self, u, dev_path, dart=None):
        node = u.adt[dev_path]
        asc_base = node.get_reg(0)[0]
        AOPBase.__init__(self, u)
        super().__init__(u, asc_base, dart)
        self.dart = dart



# SPDX-License-Identifier: MIT
import struct
from collections import defaultdict, deque
from enum import IntEnum

from ..trace.asc import ASCRegs
from ..utils import *


class BootRomMsg(IntEnum):
    GET_STATUS = 2
    BOOT_TZ0 = 5
    BOOT_IMG4 = 6
    SET_SHMEM = 0x18


class BootRomStatus(IntEnum):
    STATUS_OK = 0x66
    STATUS_BOOT_TZ0_DONE = 0x69
    STATUS_BOOT_IMG4_DONE = 0x6A
    STATUS_BOOT_UNK_DONE = 0xD2


class SEPMessage(Register64):
    EP = 7, 0
    TAG = 15, 8
    TYPE = 23, 16
    PARAM = 31, 24
    DATA = 63, 32


# TODO: make this class actually own the shared memory instead of just
#       generating a static buffer if we actually need to read/write to
#       individual items inside the shmem buffer
class SEPShMem:
    def __init__(self):
        self.items = []
        self.offset = 0x4000

    def add_item(self, name, data, min_size=0):
        sz = align_up(len(data) + 4, 0x4000)
        sz = max(sz, min_size)
        self.items.append((name, self.offset, sz, struct.pack("<I", len(data)) + data))
        self.offset += sz

    def finalize(self):
        bfr = bytearray(b"\x00" * self.offset)
        for i, (name, offset, sz, data) in enumerate(self.items):
            bfr[i * 16 : i * 16 + 12] = struct.pack("<4sII", name, sz, offset)
            bfr[offset : offset + len(data)] = data

        cnt = len(self.items)
        bfr[cnt * 16 : cnt * 16 + 4] = b"llun" # null

        return bfr


class SEP:
    SHMEM_IOVA = 0xBEEF0000
    FW_IOVA = 0xDEAD0000

    def __init__(self, proxy, iface, utils):
        self.i = iface
        self.p = proxy
        self.u = utils

        self.sep_base = self.u.adt["/arm-io/sep"].get_reg(0)[0]
        self.dart_base = self.u.adt["/arm-io/dart-sep"].get_reg(0)[0]

        self.asc = ASCRegs(self.u, self.sep_base)

        self.dart_handle = self.p.dart_init(self.dart_base, 0)

        self.epnum2name = {}
        self.epname2num = {}
        self.msgs = defaultdict(deque)

    def map_sepfw(self):
        sepfw_addr, sepfw_size = self.u.adt["/chosen/memory-map"].SEPFW
        self.p.dart_map(self.dart_handle, self.FW_IOVA, sepfw_addr, sepfw_size)

    def unmap_sepfw(self):
        _, sepfw_size = self.u.adt["/chosen/memory-map"].SEPFW
        self.p.dart_unmap(self.dart_handle, self.FW_IOVA, sepfw_size)

    def create_shmem(self):
        shmem = SEPShMem()

        # PNIC - panic buffer
        shmem.add_item(b"CINP", b"\x00", 0x8000)

        # ALPO / SIPS - unknown img4-like blobs from the ADT
        addr, sz = self.u.adt["/chosen/boot-object-manifests"].lpol
        shmem.add_item(b"OPLA", self.i.readmem(addr, sz))
        addr, sz = self.u.adt["/chosen/boot-object-manifests"].ibot
        shmem.add_item(b"IPIS", self.i.readmem(addr, sz))

        bfr = shmem.finalize()
        sz = align_up(len(bfr), 0x4000)
        self.shmem = self.u.heap.memalign(0x4000, 0x30000)
        self.i.writemem(self.shmem, bfr)
        self.p.dart_map(self.dart_handle, self.SHMEM_IOVA, self.shmem, 0x30000)

    def boot(self):
        self.create_shmem()
        self.map_sepfw()

        self.send_msg(SEPMessage(EP=0xFF, TYPE=BootRomMsg.GET_STATUS))
        self.expect_msg(0xFF, BootRomStatus.STATUS_OK)

        self.send_msg(SEPMessage(EP=0xFF, TYPE=BootRomMsg.BOOT_TZ0))
        self.expect_msg(0xFF, BootRomStatus.STATUS_BOOT_TZ0_DONE)
        self.expect_msg(0xFF, BootRomStatus.STATUS_BOOT_UNK_DONE)

        self.send_msg(SEPMessage(EP=0xFF, TYPE=BootRomMsg.GET_STATUS))
        self.expect_msg(0xFF, BootRomStatus.STATUS_OK)

        self.send_msg(
            SEPMessage(EP=0xFF, TYPE=BootRomMsg.BOOT_IMG4, DATA=self.FW_IOVA >> 0xC)
        )
        self.send_msg(
            SEPMessage(EP=0xFE, TYPE=BootRomMsg.SET_SHMEM, DATA=self.SHMEM_IOVA >> 0xC)
        )

        self.expect_msg(0xFF, BootRomStatus.STATUS_BOOT_IMG4_DONE)

        self.unmap_sepfw()

    def expect_msg(self, ep, type):
        msg = self.recv_msg(ep, block=True)
        if msg.TYPE != type:
            raise ValueError(
                f"Expected type 0x{type:x} but got message with type 0x{msg.TYPE:x}"
            )

    def send_msg(self, msg):
        self.asc.INBOX0 = msg.value
        self.asc.INBOX1 = 0

    def _recv_single_msg(self):
        msg = SEPMessage(self.asc.OUTBOX0.val)
        _ = self.asc.OUTBOX1.val
        return msg

    def _try_recv_msgs(self):
        while not self.asc.OUTBOX_CTRL.reg.EMPTY:
            msg = self._recv_single_msg()
            self.msgs[msg.EP].append(msg)
        self._handle_ep_discovery()

    def _handle_ep_discovery(self):
        while len(self.msgs[0xFD]):
            msg = self.msgs[0xFD].popleft()
            if msg.TYPE == 0:
                cs = "".join(
                    [chr((msg.DATA >> (i * 8)) & 0xFF) for i in range(3, -1, -1)]
                )
                self.epnum2name[msg.PARAM] = cs
                self.epname2num[cs] = msg.PARAM

    def recv_msg(self, ep, block=False):
        self._try_recv_msgs()
        while block and len(self.msgs[ep]) < 1:
            self._try_recv_msgs()

        if len(self.msgs[ep]):
            return self.msgs[ep].popleft()
        else:
            return None
